import bionetgen as bng
import os, re

from bionetgen.main import BioNetGen
from .utils import find_BNG_path, run_command, ActionList
from tempfile import TemporaryDirectory

# This allows access to the CLIs config setup
app = BioNetGen()
app.setup()
conf = app.config["bionetgen"]
def_bng_path = conf["bngpath"]


class BNGFile:
    """
    File object designed to deal with .bngl file manipulations.

    Usage: BNGFile(bngl_path)
           BNGFile(bngl_path, BNGPATH)

    Attributes
    ----------
    path : str
        path to the file the object needs to deal with
    _action_list : list[str]
        list of acceptible actions
    BNGPATH : str
        optional path to bng folder that contains BNG2.pl
    bngexec : str
        path to BNG2.pl

    Methods
    -------
    generate_xml(xml_file, model_file=None) : bool
        takes the given BNGL file and generates a BNG-XML from it
    strip_actions(model_path, folder) : str
        deletes actions from a given BNGL file
    write_xml(open_file, xml_type="bngxml", bngl_str=None) : bool
        given a bngl file or a string, writes an SBML or BNG-XML from it
    """

    def __init__(self, path, BNGPATH=def_bng_path) -> None:
        self.path = path
        AList = ActionList()
        self._action_list = [i + "(" for i in AList.possible_types]
        BNGPATH, bngexec = find_BNG_path(BNGPATH)
        self.BNGPATH = BNGPATH
        self.bngexec = bngexec
        self.parsed_actions = []

    def generate_xml(self, xml_file, model_file=None) -> bool:
        """
        generates an BNG-XML file from a given model file. Defaults
        to self.path if model_file is not given
        """
        if model_file is None:
            model_file = self.path
        cur_dir = os.getcwd()
        # temporary folder to work in
        with TemporaryDirectory() as temp_folder:
            # make a stripped copy without actions in the folder
            stripped_bngl = self.strip_actions(model_file, temp_folder)
            # run with --xml
            os.chdir(temp_folder)
            # TODO: take stdout option from app instead
            rc, _ = run_command(["perl", self.bngexec, "--xml", stripped_bngl])
            if rc == 1:
                # if we fail, print out what we have to
                # let the user know what BNG2.pl says
                # if rc.stdout is not None:
                #     print(rc.stdout.decode('utf-8'))
                # if rc.stderr is not None:
                #     print(rc.stderr.decode('utf-8'))
                # go back to our original location
                os.chdir(cur_dir)
                # shutil.rmtree(temp_folder)
                return False
            else:
                # we should now have the XML file
                path, model_name = os.path.split(stripped_bngl)
                model_name = model_name.replace(".bngl", "")
                written_xml_file = model_name + ".xml"
                with open(written_xml_file, "r", encoding="UTF-8") as f:
                    content = f.read()
                    xml_file.write(content)
                # since this is an open file, to read it later
                # we need to go back to the beginning
                xml_file.seek(0)
                # go back to our original location
                os.chdir(cur_dir)
                return True

    def strip_actions(self, model_path, folder) -> str:
        """
        Strips actions from a BNGL file and makes a copy
        into the given folder
        """
        # Get model name and setup path stuff
        path, model_file = os.path.split(model_path)
        # open model and strip actions
        with open(model_path, "r", encoding="UTF-8") as mf:
            # read and strip actions
            mstr = mf.read()
            # TODO: Clean this up _a lot_
            # this removes any new line escapes (\ \n) to continue
            # to another line, so we can just remove the action lines
            mstr = re.sub(r"\\\n", "", mstr)
            mlines = mstr.split("\n")
            stripped_lines = list(filter(lambda x: self._not_action(x), mlines))
            # remove spaces, actions don't allow them
            self.parsed_actions = [
                x.replace(" ", "")
                for x in filter(lambda x: not self._not_action(x), mlines)
            ]
            # let's remove begin/end actions, rarely used but should be removed
            remove_from = -1
            remove_to = -1
            for iline, line in enumerate(stripped_lines):
                if re.match(r"\s*(begin)\s+(actions)\s*", line):
                    remove_from = iline
                elif re.match(r"\s*(end)\s+(actions)\s*", line):
                    remove_to = iline
            if remove_from > 0:
                # we have a begin/end actions block
                if remove_to < 0:
                    raise RuntimeError(
                        f'There is a "begin actions" statement at line {remove_from} without a matching "end actions" statement'
                    )
                stripped_lines = (
                    stripped_lines[:remove_from] + stripped_lines[remove_to + 1 :]
                )
            if remove_to > 0:
                if remove_from < 0:
                    raise RuntimeError(
                        f'There is an "end actions" statement at line {remove_to} without a matching "begin actions" statement'
                    )
        # TODO: read stripped lines and store the actions
        # open new file and write just the model
        stripped_model = os.path.join(folder, model_file)
        stripped_lines = [x + "\n" for x in stripped_lines]
        with open(stripped_model, "w", encoding="UTF-8") as sf:
            sf.writelines(stripped_lines)
        return stripped_model

    def _not_action(self, line) -> bool:
        for action in self._action_list:
            if action in line:
                return False
        return True

    def write_xml(self, open_file, xml_type="bngxml", bngl_str=None) -> bool:
        """
        write new BNG-XML or SBML of file by calling BNG2.pl again
        or can take BNGL string in as well.
        """
        # TODO: Implement the route where this function uses the file itself
        # for this generation
        if bngl_str is None:
            # should load in the right str here
            raise NotImplementedError

        cur_dir = os.getcwd()
        # temporary folder to work in
        with TemporaryDirectory() as temp_folder:
            # write the current model to temp folder
            os.chdir(temp_folder)
            with open("temp.bngl", "w", encoding="UTF-8") as f:
                f.write(bngl_str)
            # run with --xml
            # TODO: Make output supression an option somewhere
            if xml_type == "bngxml":
                rc, _ = run_command(["perl", self.bngexec, "--xml", "temp.bngl"])
                if rc == 1:
                    print("XML generation failed")
                    # go back to our original location
                    os.chdir(cur_dir)
                    return False
                else:
                    # we should now have the XML file
                    with open("temp.xml", "r", encoding="UTF-8") as f:
                        content = f.read()
                        open_file.write(content)
                    # go back to beginning
                    open_file.seek(0)
                    os.chdir(cur_dir)
                    return True
            elif xml_type == "sbml":
                command = ["perl", self.bngexec, "temp.bngl"]
                rc, _ = run_command(command)
                if rc == 1:
                    print("SBML generation failed")
                    # go back to our original location
                    os.chdir(cur_dir)
                    return False
                else:
                    # we should now have the SBML file
                    with open("temp_sbml.xml", "r", encoding="UTF-8") as f:
                        content = f.read()
                        open_file.write(content)
                    open_file.seek(0)
                    os.chdir(cur_dir)
                    return True
            else:
                print("XML type {} not recognized".format(xml_type))
                return False



# -*- coding: utf-8 -*-
"""
Created on Fri Mar  1 16:14:42 2013

@author: proto
"""

#!/usr/bin/env python
from collections import OrderedDict
import time
import libsbml
import bionetgen.atomizer.writer.bnglWriter as writer
from optparse import OptionParser
import bionetgen.atomizer.atomizer.moleculeCreation as mc
import sys
from os import listdir
import re
import pickle
import copy

log = {"species": [], "reactions": []}
from collections import Counter, namedtuple

import bionetgen.atomizer.utils.structures as structures
from bionetgen.atomizer.utils.util import (
    logMess,
    setupLog,
    setupStreamLog,
    finishStreamLog,
    TranslationException,
)
from bionetgen.atomizer.utils import consoleCommands
from bionetgen.atomizer.sbml2bngl import SBML2BNGL

# from biogrid import loadBioGridDict as loadBioGrid
import logging
from bionetgen.atomizer.rulifier import postAnalysis
import pprint
import fnmatch
from collections import defaultdict

import sympy
from sympy.printing.str import StrPrinter
from sympy.core.sympify import SympifyError

# returntype for the sbml analyzer translator and helper functions
AnalysisResults = namedtuple(
    "AnalysisResults",
    [
        "rlength",
        "slength",
        "reval",
        "reval2",
        "clength",
        "rdf",
        "finalString",
        "speciesDict",
        "database",
        "annotation",
    ],
)


def loadBioGrid():
    pass


def handler(signum, frame):
    print("Forever is over!")
    raise Exception("end of time")


def getFiles(directory, extension):
    """
    Gets a list of bngl files that could be correctly translated in a given 'directory'

    Keyword arguments:
    directory -- The directory we will recurseviley get files from
    extension -- A file extension filter
    """
    matches = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, "*.{0}".format(extension)):
            filepath = os.path.abspath(os.path.join(root, filename))
            matches.append([filepath, os.path.getsize(os.path.join(root, filename))])

    # sort by size
    # matches.sort(key=lambda filename: filename[1], reverse=False)

    matches = [x[0] for x in matches]

    return matches


import os.path


def resource_path(relative_path):
    """Get absolute path to resource, works for dev and for PyInstaller"""
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")

    return os.path.join(base_path, relative_path)


def evaluation(numMolecules, translator):
    originalElements = numMolecules
    nonStructuredElements = len([1 for x in translator if "()" in str(translator[x])])
    if originalElements > 0:
        ruleElements = (
            (len(translator) - nonStructuredElements) * 1.0 / originalElements
        )
        if ruleElements > 1:
            ruleElements = (
                (len(translator) - nonStructuredElements) * 1.0 / len(translator.keys())
            )

    else:
        ruleElements = 0
    return ruleElements


def selectReactionDefinitions(bioNumber):
    """
    This method rrough the stats-biomodels database looking for the
    best reactionDefinitions definition available
    """
    # with open('stats4.npy') as f:
    #    db = pickle.load(f)
    fileName = resource_path("config/reactionDefinitions.json")
    useID = True
    naming = resource_path("config/namingConventions.json")
    """
    for element in db:
        if element[0] == bioNumber and element[1] != '0':
            fileName = 'reactionDefinitions/reactionDefinition' + element[1] + '.json'
            useID = element[5]
        elif element[0] > bioNumber:
            break
    """
    return fileName, useID, naming


def resolveDependencies(dictionary, key, idx):
    counter = 0
    for element in dictionary[key]:
        if idx < 20:
            counter += resolveDependencies(dictionary, element, idx + 1)
        else:
            counter += 1
    return len(dictionary[key]) + counter


def validateReactionUsage(reactant, reactions):
    for element in reactions:
        if reactant in element:
            return element
    return None


def readFromString(
    inputString,
    reactionDefinitions,
    useID,
    speciesEquivalence=None,
    atomize=False,
    loggingStream=None,
    replaceLocParams=True,
    obs_map_file=None,
):
    """
    one of the library's main entry methods. Process data from a string
    """

    console = None
    if loggingStream:
        console = logging.StreamHandler(loggingStream)
        console.setLevel(logging.DEBUG)

        setupStreamLog(console)

    reader = libsbml.SBMLReader()
    document = reader.readSBMLFromString(inputString)
    parser = SBML2BNGL(
        document.getModel(),
        useID,
        replaceLocParams=replaceLocParams,
        obs_map_file=obs_map_file,
    )

    bioGrid = False
    pathwaycommons = True
    if bioGrid:
        loadBioGrid()

    database = structures.Databases()
    database.assumptions = defaultdict(set)
    database.document = document
    database.forceModificationFlag = True
    database.reactionDefinitions = reactionDefinitions
    database.useID = useID
    database.atomize = atomize
    database.speciesEquivalence = speciesEquivalence
    database.pathwaycommons = True
    database.isConversion = True
    # if pathwaycommons:
    #    database.pathwaycommons = True
    namingConventions = resource_path("config/namingConventions.json")

    if atomize:
        translator, onlySynDec = mc.transformMolecules(
            parser,
            database,
            reactionDefinitions,
            namingConventions,
            speciesEquivalence,
            bioGrid,
        )
        database.species = translator.keys()
    else:
        translator = {}
    # logging.getLogger().flush()
    if loggingStream:
        finishStreamLog(console)
    returnArray = analyzeHelper(
        document,
        reactionDefinitions,
        useID,
        "",
        speciesEquivalence,
        atomize,
        translator,
        database,
        replaceLocParams=replaceLocParams,
        obs_map_file=obs_map_file,
    )

    if atomize and onlySynDec:
        returnArray = list(returnArray)
    returnArray = AnalysisResults(
        *(list(returnArray[0:-2]) + [database] + [returnArray[-1]])
    )

    return returnArray


def processFunctions(functions, sbmlfunctions, artificialObservables, tfunc):
    """
    this method goes through the list of functions and removes all
    sbml elements that are extraneous to bngl
    """
    # reformat time function
    for idx in range(0, len(functions)):
        """
        remove calls to functions inside functions
        """
        modificationFlag = True
        recursionIndex = 0
        # remove calls to other sbml functions
        while modificationFlag and recursionIndex < 20:
            modificationFlag = False
            for sbml in sbmlfunctions:
                if sbml in functions[idx]:
                    temp = writer.extendFunction(
                        functions[idx], sbml, sbmlfunctions[sbml]
                    )
                    if temp != functions[idx]:
                        functions[idx] = temp
                        modificationFlag = True
                        recursionIndex += 1
                        break

        functions[idx] = re.sub(r"(\W|^)(time)(\W|$)", r"\1time()\3", functions[idx])
        functions[idx] = re.sub(r"(\W|^)(Time)(\W|$)", r"\1time()\3", functions[idx])
        functions[idx] = re.sub(r"(\W|^)(t)(\W|$)", r"\1time()\3", functions[idx])

        # remove true and false
        functions[idx] = re.sub(r"(\W|^)(true)(\W|$)", r"\1 1\3", functions[idx])
        functions[idx] = re.sub(r"(\W|^)(false)(\W|$)", r"\1 0\3", functions[idx])
    # functions.extend(sbmlfunctions)
    dependencies2 = {}
    for idx in range(0, len(functions)):
        dependencies2[functions[idx].split(" = ")[0].split("(")[0].strip()] = []
        for key in artificialObservables:
            oldfunc = functions[idx]
            functions[idx] = re.sub(
                r"(\W|^)({0})([^\w(]|$)".format(key), r"\1\2()\3", functions[idx]
            )
            if oldfunc != functions[idx]:
                dependencies2[functions[idx].split(" = ")[0].split("(")[0]].append(key)
        for element in sbmlfunctions:
            oldfunc = functions[idx]
            key = element.split(" = ")[0].split("(")[0]
            if (
                re.search("(\W|^){0}(\W|$)".format(key), functions[idx].split(" = ")[1])
                != None
            ):
                dependencies2[functions[idx].split(" = ")[0].split("(")[0]].append(key)
        for element in tfunc:
            key = element.split(" = ")[0].split("(")[0]
            if key in functions[idx].split(" = ")[1]:
                dependencies2[functions[idx].split(" = ")[0].split("(")[0]].append(key)
    """
    for counter in range(0, 3):
        for element in dependencies2:
            if len(dependencies2[element]) > counter:
                dependencies2[element].extend(dependencies2[dependencies2[element][counter]])
    """

    fd = []
    for function in functions:
        # print(function, '---', dependencies2[function.split(' = ' )[0].split('(')[0]], '---', function.split(' = ' )[0].split('(')[0], 0)
        fd.append(
            [
                function,
                resolveDependencies(
                    dependencies2, function.split(" = ")[0].split("(")[0], 0
                ),
            ]
        )
    fd = sorted(fd, key=lambda rule: rule[1])
    functions = [x[0] for x in fd]

    return functions


def extractAtoms(species):
    """
    given a list of structures, returns a list
    of individual molecules/compartment pairs
    appends a number for
    """
    listOfAtoms = set()
    for molecule in species.molecules:
        for component in molecule.components:
            listOfAtoms.add(tuple([molecule.name, component.name]))
    return listOfAtoms


def bondPartners(species, bondNumber):
    relevantComponents = []
    for molecule in species.molecules:
        for component in molecule.components:
            if bondNumber in component.bonds:
                relevantComponents.append(tuple([molecule.name, component.name]))
    return relevantComponents


def getMoleculeByName(species, atom):
    """
    returns the state of molecule-component contained in atom
    """

    stateVectorVector = []
    for molecule in species.molecules:
        if molecule.name == atom[0]:
            stateVector = []
            for component in molecule.components:
                if component.name == atom[1]:

                    # get whatever species this atom is bound to
                    if len(component.bonds) > 0:
                        comp = bondPartners(species, component.bonds[0])
                        comp.remove(atom)
                        if len(comp) > 0:
                            stateVector.append(comp[0])
                        else:
                            stateVector.append("")
                    else:
                        stateVector.append("")
                    if len(component.states) > 0:
                        stateVector.append(component.activeState)
                    else:
                        stateVector.append("")
            stateVectorVector.append(stateVector)
    return tuple(stateVectorVector[0])


def extractCompartmentCoIncidence(species):
    atomPairDictionary = {}
    if [x.name for x in species.molecules] == ["EGF", "EGF", "EGFR", "EGFR"]:
        pass
    for molecule in species.molecules:
        for component in molecule.components:
            for component2 in molecule.components:
                if component == component2:
                    continue
                atom = tuple([molecule.name, component.name])
                atom2 = tuple([molecule.name, component2.name])
                molId1 = getMoleculeByName(species, atom)
                molId2 = getMoleculeByName(species, atom2)
                key = tuple([atom, atom2])
                # print(key, (molId1, molId2))
                if key not in atomPairDictionary:
                    atomPairDictionary[key] = Counter()
                atomPairDictionary[key].update([tuple([molId1, molId2])])

    return atomPairDictionary


def extractCompartmentStatistics(
    bioNumber, useID, reactionDefinitions, speciesEquivalence
):
    """
    Iterate over the translated species and check which compartments
    are used together, and how.
    """
    reader = libsbml.SBMLReader()
    document = reader.readSBMLFromFile(bioNumber)

    parser = SBML2BNGL(document.getModel(), useID)
    database = structures.Databases()
    database.pathwaycommons = False
    # call the atomizer (or not)
    # if atomize:
    translator, onlySynDec = mc.transformMolecules(
        parser, database, reactionDefinitions, speciesEquivalence
    )
    # else:
    #    translator={}

    compartmentPairs = {}
    for element in translator:
        temp = extractCompartmentCoIncidence(translator[element])
        for element in temp:
            if element not in compartmentPairs:
                compartmentPairs[element] = temp[element]
            else:
                compartmentPairs[element].update(temp[element])
    finalCompartmentPairs = {}
    print("-----")
    for element in compartmentPairs:
        if element[0][0] not in finalCompartmentPairs:
            finalCompartmentPairs[element[0][0]] = {}
        finalCompartmentPairs[element[0][0]][
            tuple([element[0][1], element[1][1]])
        ] = compartmentPairs[element]
    return finalCompartmentPairs


def recursiveSearch(dictionary, element, visitedFunctions=[]):
    tmp = 0
    for item in dictionary[element]:
        if dictionary[item] == []:
            tmp += 1
        else:
            if item in visitedFunctions:
                raise Exception(
                    "Recursive function search landed twice in the same function"
                )
            tmp += 1
            tmp += recursiveSearch(dictionary, item, [item] + visitedFunctions)
    return tmp


def reorder_and_replace_arules(functions, parser):
    # TODO: Check if we need full_prec, make it optional
    prnter = StrPrinter({"full_prec": False})
    func_names = []
    # get full func names and initialize dependency graph
    dep_dict = {}
    for func in functions:
        splt = func.split("=")
        n = splt[0]
        f = "=".join(splt[1:])
        name = n.rstrip().replace("()", "")
        func_names.append(name)
        if "fRate" not in name:
            dep_dict[name] = []
    # make dependency graph between funcs only
    func_dict = {}
    new_funcs = []
    # Let's replace and build dependency map
    frates = []
    for func in functions:
        splt = func.split("=")
        # TODO: turn this into warning
        n = splt[0]
        f = "=".join(splt[1:])
        fname = n.rstrip().replace("()", "")
        try:
            fs = sympy.sympify(f, locals=parser.all_syms)
        except:
            # Can't parse this func
            if fname.startswith("fRate"):
                frates.append((fname.strip(), f))
            else:
                func_dict[fname] = f
            continue
        # replace here since it affects dependency
        for item in parser.only_assignment_dict.items():
            oname, nname = item
            osym, ns = sympy.symbols(oname + "," + nname)
            fs = fs.subs(osym, ns)
        func_dict[fname] = fs
        # need to build a dependency graph to figure out what to
        # write first
        # We can skip this if it's a functionRate
        if "fRate" not in n:
            list_of_deps = list(map(str, fs.atoms(sympy.Symbol)))
            for dep in list_of_deps:
                if dep in func_names:
                    dep_dict[fname].append(dep)
        else:
            frates.append((n.strip(), fs))
    # Now reorder accordingly
    ordered_funcs = []
    # this ensures we write the independendent functions first
    stck = sorted(dep_dict.keys(), key=lambda x: len(dep_dict[x]))
    # FIXME: This algorithm works but likely inefficient
    while len(stck) > 0:
        k = stck.pop()
        deps = dep_dict[k]
        if len(deps) == 0:
            if k not in ordered_funcs:
                ordered_funcs.append(k)
        else:
            stck.append(k)
            for dep in deps:
                if dep not in ordered_funcs:
                    stck.append(dep)
                dep_dict[k].remove(dep)
    # print ordered functions and return
    for fname in ordered_funcs:
        fs = func_dict[fname]
        # clean up some whitespace here
        try:
            func_str = prnter.doprint(fs)
            func_str = func_str.replace("**", "^")
            new_funcs.append(fname + "() = " + func_str)
        except:
            new_funcs.append(fname + "() = " + fs)
    # now write the functionRates
    for fitem in frates:
        n, fs = fitem
        func_str = prnter.doprint(fs)
        func_str = func_str.replace("**", "^")
        new_funcs.append(n + " = " + func_str)
    return new_funcs


def reorderFunctions(functions):
    """
    Analyze a list of sbml functions and make sure there are no forward dependencies.
    Reorder if necessary
    """
    functionNames = []
    tmp = []
    for function in functions:

        m = re.split("(?<=\()[\w)]", function)
        functionName = m[0]
        if "=" in functionName:
            functionName = functionName.split("=")[0].strip() + "("
        functionNames.append(functionName)
    functionNamesDict = {x: [] for x in functionNames}
    for idx, function in enumerate(functions):
        tmp = [
            x
            for x in functionNames
            if x in function.split("=")[1] and x != functionNames[idx]
        ]
        functionNamesDict[functionNames[idx]].extend(tmp)
    newFunctionNamesDict = {}
    for name in functionNamesDict:
        try:
            newFunctionNamesDict[name] = recursiveSearch(functionNamesDict, name, [])
        # there is a circular dependency
        except:
            newFunctionNamesDict[name] = 99999
    functionWeightsDict = {x: newFunctionNamesDict[x] for x in newFunctionNamesDict}
    functionWeights = []
    for name in functionNames:
        functionWeights.append(functionWeightsDict[name])
    tmp = zip(functions, functionWeights)
    idx = sorted(tmp, key=lambda x: x[1])
    return [x[0] for x in idx]


def postAnalysisHelper(outputFile, bngLocation, database):
    consoleCommands.setBngExecutable(bngLocation)
    outputDir = os.sep.join(outputFile.split(os.sep)[:-1])
    if outputDir != "":
        retval = os.getcwd()
        os.chdir(outputDir)
    consoleCommands.bngl2xml(outputFile.split(os.sep)[-1])
    if outputDir != "":
        os.chdir(retval)
    bngxmlFile = ".".join(outputFile.split(".")[:-1]) + ".xml"
    # print('Sending BNG-XML file to context analysis engine')
    contextAnalysis = postAnalysis.ModelLearning(bngxmlFile)
    # analysis of redundant bonds
    deleteBonds = contextAnalysis.analyzeRedundantBonds(
        database.assumptions["redundantBonds"]
    )
    for molecule in database.assumptions["redundantBondsMolecules"]:
        if molecule[0] in deleteBonds:
            for bond in deleteBonds[molecule[0]]:
                database.translator[molecule[1]].deleteBond(bond)
                logMess(
                    "INFO:CTX002",
                    "Used context information to determine that the bond {0} in species {1} is not likely".format(
                        bond, molecule[1]
                    ),
                )


def postAnalyzeFile(
    outputFile, bngLocation, database, replaceLocParams=True, obs_map_file=None
):
    """
    Performs a postcreation file analysis based on context information
    """
    # print('Transforming generated BNG file to BNG-XML representation for analysis')

    postAnalysisHelper(outputFile, bngLocation, database)

    # recreate file using information from the post analysis
    returnArray = analyzeHelper(
        database.document,
        database.reactionDefinitions,
        database.useID,
        outputFile,
        database.speciesEquivalence,
        database.atomize,
        database.translator,
        database,
        replaceLocParams=replaceLocParams,
        obs_map_file=obs_map_file,
    )
    with open(outputFile, "w", encoding="UTF-8") as f:
        f.write(returnArray.finalString)
    # recompute bng-xml file
    consoleCommands.bngl2xml(outputFile)
    bngxmlFile = ".".join(outputFile.split(".")[:-1]) + ".xml"
    # recompute context information
    contextAnalysis = postAnalysis.ModelLearning(bngxmlFile)

    # get those species patterns that follow uncommon motifs
    motifSpecies, motifDefinitions = contextAnalysis.processContextMotifInformation(
        database.assumptions["lexicalVsstoch"], database
    )
    # motifSpecies, motifDefinitions = contextAnalysis.processAllContextInformation()
    if len(motifDefinitions) > 0:
        logMess(
            "INFO:CTX003",
            "Species with suspect context information were found. Information is being dumped to {0}_context.log".format(
                outputFile
            ),
        )
        with open("{0}_context.log".format(outputFile), "w") as f:
            pprint.pprint(dict(motifSpecies), stream=f)
            pprint.pprint(motifDefinitions, stream=f)

    # score hypothetical bonds
    # contextAnalysis.scoreHypotheticalBonds(assumptions['unknownBond'])


def postAnalyzeString(outputFile, bngLocation, database):
    postAnalysisHelper(outputFile, bngLocation, database)

    # recreate file using information from the post analysis
    returnArray = analyzeHelper(
        database.document,
        database.reactionDefinitions,
        database.useID,
        outputFile,
        database.speciesEquivalence,
        database.atomize,
        database.translator,
        database,
    ).finalString
    return returnArray


def analyzeFile(
    bioNumber,
    reactionDefinitions,
    useID,
    namingConventions,
    outputFile,
    speciesEquivalence=None,
    atomize=False,
    bioGrid=False,
    pathwaycommons=False,
    ignore=False,
    noConversion=False,
    memoizedResolver=True,
    replaceLocParams=True,
    quietMode=False,
    logLevel="WARNING",
    obs_map_file=None,
):
    """
    one of the library's main entry methods. Process data from a file
    """
    """
    import cProfile, pstats, StringIO
    pr = cProfile.Profile()
    pr.enable()
    """
    setupLog(
        outputFile + ".log", getattr(logging, logLevel.upper()), quietMode=quietMode
    )

    logMess.log = []
    logMess.counter = -1
    reader = libsbml.SBMLReader()
    document = reader.readSBMLFromFile(bioNumber)

    if document.getModel() == None:
        print(
            "l714 - File {0} could not be recognized as a valid SBML file".format(
                bioNumber
            )
        )
        return
    parser = SBML2BNGL(
        document.getModel(),
        useID,
        replaceLocParams=replaceLocParams,
        obs_map_file=obs_map_file,
    )
    parser.setConversion(not noConversion)
    database = structures.Databases()
    database.assumptions = defaultdict(set)
    database.forceModificationFlag = True
    database.pathwaycommons = pathwaycommons
    database.ignore = ignore
    database.assumptions = defaultdict(set)

    bioGridDict = {}
    if bioGrid:
        bioGridDict = loadBioGrid()

    # call the atomizer (or not). structured molecules are contained in translator
    # onlysyndec is a boolean saying if a model is just synthesis of decay reactions
    # ASS2019 - With this try/except the translator was not being initialized and led to an undefined error in certain models

    translator = {}
    try:
        if atomize:
            translator, onlySynDec = mc.transformMolecules(
                parser,
                database,
                reactionDefinitions,
                namingConventions,
                speciesEquivalence,
                bioGrid,
                memoizedResolver,
            )
    except TranslationException as e:
        print(
            "Found an error in {0}. Check log for more details. Use -I to ignore translation errors".format(
                e.value
            )
        )
        if len(logMess.log) > 0:
            with open(outputFile + ".log", "w") as f:
                for element in logMess.log:
                    f.write(element + "\n")
        return

    # process other sections of the sbml file (functions reactions etc.)
    """
    pr.disable()
    s = StringIO.StringIO()
    sortby = 'cumulative'
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats(10)
    print(s.getvalue())
    """
    database.document = document
    database.reactionDefinitions = reactionDefinitions
    database.useID = useID
    database.speciesEquivalence = speciesEquivalence
    database.atomize = atomize
    database.isConversion = not noConversion

    returnArray = analyzeHelper(
        document,
        reactionDefinitions,
        useID,
        outputFile,
        speciesEquivalence,
        atomize,
        translator,
        database,
        replaceLocParams=replaceLocParams,
        obs_map_file=obs_map_file,
    )

    with open(outputFile, "w", encoding="UTF-8") as f:
        f.write(returnArray.finalString)
    # with open('{0}.dict'.format(outputFile), 'wb') as f:
    #    pickle.dump(returnArray[-1], f)
    if atomize and onlySynDec:
        returnArray = list(returnArray)
        # returnArray.translator = -1
    returnArray = AnalysisResults(
        *(list(returnArray[0:-2]) + [database] + [returnArray[-1]])
    )
    return returnArray


def correctRulesWithParenthesis(rules, parameters):
    """
    helper function. Goes through a list of rules and adds a parenthesis
    to the reaction rates of those functions whose rate is in list
    'parameters'.
    """
    for idx in range(len(rules)):
        tmp = [x for x in parameters if x + " " in rules[idx]]
        # for tmpparameter in tmp:
        #    re.sub(r'(\W|^){0}(\W|$)'.format(tmpparameter), r'\1{0}\2'.format(dictionary[key]), tmp[1])
        if len(tmp) > 0:
            rules[idx].strip()
            rules[idx] += "()"


def changeNames(functions, dictionary):
    """
    changes instances of keys in dictionary appeareing in functions to their corresponding
    alternatives
    """
    tmpArray = []
    for function in functions:
        tmp = function.split(" = ")
        # hack to avoid problems with less than equal or more than equal
        # in equations

        # ASS2019 - There are cases where we have more than one equal sign e.g. "= = 0&&"
        # in an if statement, I _think_ we need to re-add the '=' we removed by doing
        # split
        tmp = [tmp[0], "=".join(tmp[1:])]
        for key in [x for x in dictionary if x in tmp[1]]:
            # ASS - if the key is equal to the value, this goes for an infinite loop
            if key == dictionary[key]:
                continue
            while re.search(r"([\W, ]|^){0}([\W, ]|$)".format(key), tmp[1]):
                tmp[1] = re.sub(
                    r"([\W, ]|^){0}([\W, ]|$)".format(key),
                    r"\1{0}\2".format(dictionary[key]),
                    tmp[1],
                )
        tmpArray.append("{0} = {1}".format(tmp[0], tmp[1]))
    return tmpArray


# ASS - We need to rename the functions if they are the same as obs
def changeDefs(functions, dictionary):
    """
    changes the names of the functions (RHS) instead of the LHS
    """
    tmpArray = []
    for function in functions:
        tmp = function.split(" = ")
        # hack to avoid problems with less than equal or more than equal
        # in equations
        tmp = [tmp[0], "".join(tmp[1:])]
        for key in [x for x in dictionary if x in tmp[0]]:
            # ASS - if the key is equal to the value, this goes for an infinite loop
            if key == dictionary[key]:
                continue
            while re.search(r"([\W, ]|^){0}([\W, ]|$)".format(key), tmp[0]):
                tmp[0] = re.sub(
                    r"([\W, ]|^){0}([\W, ]|$)".format(key),
                    r"\1{0}\2".format(dictionary[key]),
                    tmp[0],
                )
        tmpArray.append("{0} = {1}".format(tmp[0], tmp[1]))
    return tmpArray


def changeRates(reactions, dictionary):
    """
    changes instances of keys in dictionary appeareing in reaction rules to their corresponding
    alternatives
    """
    tmpArray = []
    tmp = None
    for reaction in reactions:
        cmt = None
        tmp = reaction.strip().split(" ")
        for ielem, elem in enumerate(tmp):
            if elem.startswith("#"):
                cmt = tmp[ielem:]
                tmp = tmp[:ielem]
                break
        for key in [x for x in dictionary if x in tmp[-1]]:
            tmp[-1] = re.sub(
                r"(\W|^){0}(\W|$)".format(key),
                r"\1{0}\2".format(dictionary[key]),
                tmp[-1],
            )
        if cmt is not None:
            tmp += cmt
        tmpArray.append(" ".join(tmp))
    # if tmp:
    #     if cmt is not None:
    #         tmp += cmt
    #     tmpArray.append(' '.join(tmp))
    return tmpArray


def unrollFunctions(functions):
    flag = True
    # bngl doesnt accept nested function calling
    while flag:
        dictionary = OrderedDict()
        flag = False
        for function in functions:
            tmp = function.split(" = ")
            for key in dictionary:
                if key in tmp[1]:
                    tmp[1] = re.sub(
                        r"(\W|^){0}\(\)(\W|$)".format(key),
                        r"\1({0})\2".format(dictionary[key]),
                        tmp[1],
                    )
                    flag = False
            dictionary[tmp[0].split("()")[0]] = tmp[1]
        tmp = []
        for key in dictionary:
            tmp.append("{0}() = {1}".format(key, dictionary[key]))
        functions = tmp
    return functions


def analyzeHelper(
    document,
    reactionDefinitions,
    useID,
    outputFile,
    speciesEquivalence,
    atomize,
    translator,
    database,
    bioGrid=False,
    replaceLocParams=True,
    obs_map_file=None,
):
    """
    taking the atomized dictionary and a series of data structure, this method
    does the actual string output.
    """
    useArtificialRules = False
    parser = SBML2BNGL(
        document.getModel(),
        useID,
        replaceLocParams=replaceLocParams,
        obs_map_file=obs_map_file,
    )
    # ASS: Port over other parsers? used_molecules list
    if hasattr(database, "parser"):
        parser.used_molecules.extend(database.parser.used_molecules)
    parser.setConversion(database.isConversion)
    #
    param, zparam = parser.getParameters()
    rawSpecies = {}
    for species in parser.model.getListOfSpecies():
        rawtemp = parser.getRawSpecies(species, [x.split(" ")[0] for x in param])
        rawSpecies[rawtemp["identifier"]] = rawtemp
    parser.reset()

    # parser.bngModel.translator = copy.deepcopy(translator)
    parser.bngModel.translator = translator
    (
        molecules,
        initialConditions,
        observables,
        speciesDict,
        observablesDict,
        annotationInfo,
    ) = parser.getSpecies(translator, [x.split(" ")[0] for x in param])

    # finally, adjust parameters and initial concentrations according to whatever  initialassignments say
    param, zparam, initialConditions, initCondMap = parser.getInitialAssignments(
        translator, param, zparam, molecules, initialConditions
    )

    # FIXME: this method is a mess, improve handling of assignmentrules since we can actually handle those
    (
        aParameters,
        aRules,
        nonzparam,
        artificialRules,
        removeParams,
        artificialObservables,
    ) = parser.getAssignmentRules(
        zparam, param, rawSpecies, observablesDict, translator
    )

    compartments = parser.getCompartments()
    functions = []
    assigmentRuleDefinedParameters = []

    # FIXME: We should determine if an assignment rule
    # if being used along with a reaction and ignore the
    # reaction if it is being modified by both. This will
    # likely require us to feed something from the assingment
    # rule result into the following function
    reactionParameters, rules, rateFunctions = parser.getReactions(
        translator,
        len(compartments) > 1,
        atomize=atomize,
        parameterFunctions=artificialObservables,
        database=database,
    )

    functions.extend(rateFunctions)

    for element in nonzparam:
        param.append("{0} 0".format(element))
    param = [x for x in param if x not in removeParams]

    # tags = '@{0}'.format(compartments[0].split(' ')[0]) if len(compartments) == 1 else '@cell'
    # ASS - trying to remove @cell as a default compartment. Also, 0th compartment
    # is generally a comment. Here we are taking the first non-comment compartment as
    # the default compartment if we are missing the compartment, generally for
    # SBML non-constant parameter starting values.
    if len(compartments) == 0:
        def_compartment = ""
    elif len(compartments) == 1:
        def_compartment = compartments[0].split(" ")[0]
    else:
        for compartment in compartments:
            if not compartment.startswith("#"):
                def_compartment = compartment.split(" ")[0]
                break
    if def_compartment != "":
        tags = "@{0}".format(def_compartment)
    else:
        tags = ""

    # We need to replace stuff that we have a definition for
    # if they are used in assignment rules
    art_names = dict([(key[:-3], key) for key in artificialObservables])
    for key in artificialObservables:
        changed = False
        f = artificialObservables[key]

        fsplt = f.split("=")
        fn = fsplt[0]
        fd = "=".join(fsplt[1:])
        for an in art_names:
            # We need an exact match
            if re.search("\b{}\b".format(an), fd) is not None:
                fd = re.sub("\b{}\b".format(an), art_names[an], fd)
                changed = True
        if changed:
            artificialObservables[key] = fn.split()[0] + " = " + fd
    # Here we are adding removed parameters back as
    # molecules, species and observables? How do we know
    # we need these?
    for remPar in removeParams:
        par_nam = remPar.split()[0]
        write = True
        # Check assignment rules first
        for key in artificialObservables:
            if (par_nam == key) or (par_nam + "_ar" == key):
                # We have an assignment rule for this parameter
                # and we don't want to have molecules and stuff
                write = False
                break
        if write:
            if par_nam not in molecules:
                molecules.append(par_nam)
            obs_str = "Species {0} {0}".format(par_nam)
            if obs_str not in molecules:
                observables.append(obs_str)
            init_cond = par_nam + tags + " " + " ".join(remPar.split()[1:])
            if init_cond not in initialConditions:
                initialConditions.append(init_cond)
    ## Comment out those parameters that are defined with assignment rules
    ## TODO: I think this is correct, but it may need to be checked
    tmpParams = []
    for idx, parameter in enumerate(param):
        for key in artificialObservables:
            if re.search("^{0}\s".format(key), parameter) != None:
                assigmentRuleDefinedParameters.append(idx)
    tmpParams.extend(artificialObservables)
    tmpParams.extend(removeParams)
    tmpParams = set(tmpParams)
    correctRulesWithParenthesis(rules, tmpParams)

    for element in assigmentRuleDefinedParameters:
        param[element] = "#" + param[element]

    deleteMolecules = []
    deleteMoleculesFlag = True

    for key in artificialObservables:
        flag = -1
        for idx, observable in enumerate(observables):
            if "Species {0} {0}()".format(key) in observable:
                flag = idx
        if flag != -1:
            observables.pop(flag)
        functions.append(artificialObservables[key])
        flag = -1

        if "{0}()".format(key) in molecules:
            flag = molecules.index("{0}()".format(key))

        if flag != -1:
            if deleteMoleculesFlag:
                deleteMolecules.append(flag)
            else:
                deleteMolecules.append(key)
            # result =validateReactionUsage(molecules[flag], rules)
            # if result != None:
            #    logMess('ERROR', 'Pseudo observable {0} in reaction {1}'.format(molecules[flag], result))
            # molecules.pop(flag)

        flag = -1
        for idx, specie in enumerate(initialConditions):
            if ":{0}(".format(key) in specie:
                flag = idx
        if flag != -1:
            initialConditions[flag] = "#" + initialConditions[flag]

    for flag in sorted(deleteMolecules, reverse=True):

        if deleteMoleculesFlag:
            logMess(
                "WARNING:SIM101",
                "{0} reported as function, but usage is ambiguous".format(
                    molecules[flag]
                ),
            )
            result = validateReactionUsage(molecules[flag], rules)
            if result is not None:
                logMess(
                    "ERROR:Simulation",
                    "Pseudo observable {0} in reaction {1}".format(
                        molecules[flag], result
                    ),
                )

            # since we are considering it an observable delete it from the molecule and
            # initial conditions list
            # s = molecules.pop(flag)
            # initialConditions = [x for x in initialConditions if '$' + s not in x]
        else:
            logMess(
                "WARNING:SIM101",
                "{0} reported as species, but usage is ambiguous.".format(flag),
            )
            artificialObservables.pop(flag)

    sbmlfunctions = parser.getSBMLFunctions()
    functions.extend(aRules)

    parser.bngModel.sbmlFunctions = sbmlfunctions
    processFunctions(functions, sbmlfunctions, artificialObservables, rateFunctions)

    for interation in range(0, 3):
        for sbml2 in sbmlfunctions:
            for sbml in sbmlfunctions:
                if sbml == sbml2:
                    continue
                if sbml in sbmlfunctions[sbml2]:
                    sbmlfunctions[sbml2] = writer.extendFunction(
                        sbmlfunctions[sbml2], sbml, sbmlfunctions[sbml]
                    )

    # TODO: if an observable is defined via artificial obs
    # we should overwrite it in obs dict
    for key in observablesDict:
        if key + "_ar" in artificialObservables:
            observablesDict[key] = key + "_ar"
    #
    functions = reorderFunctions(functions)
    #
    functions = changeNames(functions, aParameters)
    # change reference for observables with compartment name
    functions = changeNames(functions, observablesDict)
    #
    functions = unrollFunctions(functions)
    rules = changeRates(rules, aParameters)

    ar_names = {}
    # Some observables are encoded as functions
    for func in functions:
        fname = func.split("=")[0].split("(")[0]
        if fname.endswith("_ar"):
            potential_obs = fname.replace("_ar", "")
            if potential_obs in observablesDict:
                ar_names[potential_obs] = fname
    # Switch up AR stuff that's used as rate constants
    for item in observablesDict.items():
        k, t = item
        if t.endswith("_ar"):
            ar_names[k] = t
    rules = changeRates(rules, ar_names)

    # Parameter replacement leaves a lot of unevaluated
    # math behing and it looks really ugly. I'm going
    # to parse this and try to evaluate it all

    # TODO: This needs more love, I'm definitely not
    # handling certain things I normally handle in sbml2bnl
    # using sympy, port those in or turn them into importable
    # stuff
    # TODO: Check if full_prec is bad, make it optional
    prnter = StrPrinter({"full_prec": False})
    try:
        new_funcs = []
        obs_syms = list(map(sympy.Symbol, parser.obs_names))
        for func in functions:
            splt = func.split("=")
            n = splt[0]
            f = "=".join(splt[1:])
            n, f = splt
            try:
                fs = sympy.sympify(f, locals=parser.all_syms)
            except SympifyError:
                logMess(
                    "ERROR:SYMP002",
                    "Sympy can't parse a function during post-processing",
                )
                raise TranslationException(f)
            # Test if we get a complex i from simplification
            smpl = fs.nsimplify().evalf().simplify()
            # Epsilon checking
            n, d = smpl.as_numer_denom()
            # I don't want to touch the current rate parsing so
            # I'll remove it and then add it back if needed
            # TODO: mentioned above is a temporary solution
            had_epsilon = False
            if parser.all_syms["__epsilon__"] in d.atoms():
                d = d - parser.all_syms["__epsilon__"]
                had_epsilon = True
            # for item in parser.all_syms.items():
            for s in obs_syms:
                # k, s = item
                if s in d.atoms():
                    d = d.subs(s, 0)
            if d == 0:
                if had_epsilon:
                    new_f = prnter.doprint(smpl)
                else:
                    n, d = smpl.as_numer_denom()
                    logMess(
                        "WARNING:RATE001",
                        "Post-parameter replacement, the denominator can be 0, adding an epsilon to avoid discontinuities",
                    )
                    new_f = (
                        "("
                        + prnter.doprint(n)
                        + ")/("
                        + prnter.doprint(d)
                        + "+ __epsilon__)"
                    )
                    parser.write_epsilon = True
            else:
                new_f = prnter.doprint(smpl)
            new_f = new_f.replace("**", "^")
            # We want to do this if it makes the rate constant
            # more readable
            # FIXME: This doesn't mesh well with AR replacement
            # if len(new_f) < len(func):
            #    new_funcs.append(splt[0] + " = " + new_f)
            # else:
            #    new_funcs.append(func)
            new_funcs.append(splt[0] + " = " + new_f)
        functions = new_funcs
    except:
        # raise
        # This is not essential, let's just move on if
        # sympify fails. This catch-all is here because
        # I know there will be random small things and that
        # this bit is entirely optional
        pass

    functions = reorder_and_replace_arules(functions, parser)
    # ASS2019 - we need to adjust initial conditions of assignment rules
    # so that they start with the correct values. While this doesn't
    # impact model translation quality, it does make it difficult to
    # do automated testing
    initialConditions = parser.adjustInitialConditions(
        param, initialConditions, artificialObservables, observables, functions
    )

    # ASS - We need to check for identical observables and functions. If
    # they are the same, re-number them so avoid having identical names
    # this comes up when we do non-compartmental models only really
    # due to the naming scheme of STUFF_COMPNAME but still should be
    # handled somehow
    idenObsFuncDict = {}
    for obs in observables:
        obsKey = obs.split(" ")[1]
        for aObsKey in artificialObservables.keys():
            if obsKey == aObsKey:
                idenObsFuncDict[obsKey] = obsKey + "_func"

    functions = changeDefs(functions, idenObsFuncDict)

    if len(artificialRules) + len(rules) == 0:
        logMess("ERROR:SIM203", "The file contains no reactions")
    # ASS: This "useArtificialRules" flag can never be true
    # but artificial rules are being made and left unused
    # which is incorrect, at least in some cases e.g. BioMod 207
    if useArtificialRules or len(rules) == 0:
        rules = ["#{0}".format(x) for x in rules]
        artificialRules.extend(rules)
        rules = artificialRules
    if len(artificialRules) > 0:
        logMess(
            "WARNING:ARTR001",
            "The model contains rate rules which are modeled as rules in the BNGL translation.",
        )
        rules.extend(artificialRules)
    evaluate = evaluation(len(observables), translator)
    # else:
    #    artificialRules =['#{0}'.format(x) for x in artificialRules]
    #    evaluate =  evaluation(len(observables), translator)
    #    rules.extend(artificialRules)
    commentDictionary = {}

    # ASS: removing species that are not used
    # in BNGL species do not get adjusted by functions
    # therefore we can check the rules, if a species do not
    # appear anywhere in a rule, we can remove it
    # this will clean up a lot of translations
    # TODO: My approach to observables is too simplistic
    # obs can be a whole lot more complicated, ensure what
    # I'm doing actually works
    # TODO: Some species are not used in rules and are not fixed
    # BUT are then used in some functions. The original modeller
    # should have turned these into parameters but didn't. Let's
    # turn them into parameters? or leave them be?
    # also remove from seed species
    init_to_rem = []
    turn_to_param = []
    for iss, sspec in enumerate(initialConditions):
        comp = None
        splt = sspec.split()
        sname = splt[0]
        if splt[-1].startswith("#"):
            val = " ".join(splt[1:-2])
        else:
            val = " ".join(splt[1:])
        # let's see if we have a compartment
        if "@" in sname:
            plt = sname.split(":")
            if len(plt) > 1:
                # using @comp:spec notation
                sname = plt[1]
                comp = plt[0][1:]
            else:
                # using spec@comp notation
                sname, comp = sname.split("@")
        # remove $ if it's a fixed species
        if sname.startswith("$"):
            sname = sname[1:]
        # We want only the name
        if "(" in sname:
            sname = sname[: sname.find("(")]
        # let's see if it's actually used in rules
        if sname not in parser.used_molecules:
            # this is a "fixed molecule" that doesn't get used
            # in reactions. Let's check compartment and turn
            # into parameter instead
            # TODO: Sometimes the _comp version is used, sometimes
            # not, make it consistent
            if comp is not None:
                # we have a compartment
                turn_to_param.append(sname + "_{} ".format(comp) + val)
            turn_to_param.append(sname + " " + val)
        if sname not in parser.used_molecules:
            init_to_rem.append(sspec)
    for i in init_to_rem:
        initialConditions.remove(i)
    # add any missing initial conditions to bngModel
    for icon in initialConditions:
        if icon in initCondMap:
            sym = initCondMap[icon]
            if sym in parser.bngModel.species:
                sp = parser.bngModel.species[sym]
                sp.val = " ".join(icon.split()[1:-1])
    # Turn any "fixed molecules" that are not used in rules
    # into parameters
    param.extend(turn_to_param)
    # remove from molecules
    molec_to_rem = []
    for molec in molecules:
        # name
        if "(" in molec:
            mname = molec[: molec.find("(")]
        else:
            mname = molec
        # used or not?
        if mname not in parser.used_molecules:
            molec_to_rem.append(molec)
    for i in molec_to_rem:
        molecules.remove(i)
    # and observables
    obs_to_rem = []
    for iobs, obs_str in enumerate(observables):
        oname = obs_str.split()[2]
        comp = None
        if "@" in oname:
            if len(oname.split(":")) > 1:
                # using @comp:spec
                oname = oname.split(":")[1]
            else:
                # using spec@comp
                oname, comp = oname.split("@")
        if "(" in oname:
            oname = oname[: oname.find("(")]
        if oname not in parser.used_molecules:
            obs_to_rem.append(obs_str)
    for i in obs_to_rem:
        observables.remove(i)
    # done removing useless species/seed species/obs

    # update param values
    for param_line in param:
        param_splt = param_line.split()
        param_name = param_splt[0]
        param_val = " ".join(param_splt[1:])
        if param_name in parser.bngModel.parameters:
            parser.bngModel.parameters[param_name].val = param_val

    # TODO: temporary: check structured molecule ratio
    struc_count = 0
    for molec_str in molecules:
        srch = re.search(r"(\W|^)(.+)\((.*)\)", molec_str)
        if srch is not None:
            if len(srch.group(3)) > 0:
                struc_count += 1
    struc_ratio = float(struc_count) / len(molecules) if len(molecules) > 0 else 0
    print("Structured molecule type ratio: {}".format(struc_ratio))

    # If we must, add __epsilon__ to parameter list
    if parser.write_epsilon:
        param = ["__epsilon__ 1e-100"] + param

    if atomize:
        commentDictionary[
            "notes"
        ] = "'This is an atomized translation of an SBML model created on {0}.".format(
            time.strftime("%d/%m/%Y")
        )
    else:
        commentDictionary[
            "notes"
        ] = "'This is a plain translation of an SBML model created on {0}.".format(
            time.strftime("%d/%m/%Y")
        )
    commentDictionary[
        "notes"
    ] += " The original model has {0} molecules and {1} reactions. The translated model has {2} molecules and {3} rules'".format(
        parser.model.getNumSpecies(),
        parser.model.getNumReactions(),
        len(molecules),
        len(set(rules)),
    )
    meta = parser.getMetaInformation(commentDictionary)
    finalString = writer.finalText(
        meta,
        param + reactionParameters,
        molecules,
        initialConditions,
        list(OrderedDict.fromkeys(observables)),
        list(OrderedDict.fromkeys(rules)),
        functions,
        compartments,
        annotationInfo,
        outputFile,
    )

    logMess(
        "INFO:SUM001",
        "File contains {0} molecules out of {1} original SBML species".format(
            len(molecules), len(observables)
        ),
    )

    # rate of each classified rule
    evaluate2 = 0 if len(observables) == 0 else len(molecules) * 1.0 / len(observables)

    # add unit information to annotations

    # print("in libsbml, done completely")
    # with open(outputFile, "w") as f:
    #     f.write(str(parser.bngModel))
    parser.bngModel.all_syms = parser.all_syms
    parser.bngModel.consolidate()
    finalString = str(parser.bngModel)

    annotationInfo["units"] = parser.getUnitDefinitions()
    return AnalysisResults(
        len(rules),
        len(observables),
        evaluate,
        evaluate2,
        len(compartments),
        parser.getSpeciesAnnotation(),
        finalString,
        speciesDict,
        None,
        annotationInfo,
    )

    """
    if translator != {}:
        for element in database.classifications:
            if element not in classificationDict:
                classificationDict[element] = 0.0
            classificationDict[element] += 1.0/len(database.classifications)
        return len(rules), evaluate, parser.getModelAnnotation(), classificationDict
    """
    # return None, None, None, None


def processFile(translator, parser, outputFile):
    param2 = parser.getParameters()
    molecules, species, observables, observablesDict = parser.getSpecies(translator)
    compartments = parser.getCompartments()
    param, rules, functions = parser.getReactions(translator, True)
    param += param2
    writer.finalText(
        param,
        molecules,
        species,
        observables,
        rules,
        functions,
        compartments,
        {},
        outputFile,
    )


def BNGL2XML():
    pass


def getAnnotations(annotation):
    annotationDictionary = []
    if annotation == [] or annotation is None:
        return []
    for indivAnnotation in annotation:
        for index in range(0, indivAnnotation.getNumAttributes()):
            annotationDictionary.append(indivAnnotation.getValue(index))
    return annotationDictionary


def getAnnotationsDict(annotation):
    annotationDict = {}
    for element in annotation:
        annotationDict[element] = getAnnotations(annotation[element])
    return annotationDict


def processFile2():
    for bioNumber in [19]:
        # if bioNumber in [398]:
        #    continue
        # bioNumber = 175
        logMess.log = []
        logMess.counter = -1
        reactionDefinitions, useID, naming = selectReactionDefinitions(
            "BIOMD%010i.xml" % bioNumber
        )
        print(reactionDefinitions, useID)
        # reactionDefinitions = 'reactionDefinitions/reactionDefinition7.json'
        # spEquivalence = 'reactionDefinitions/speciesEquivalence19.json'
        spEquivalence = detectCustomDefinitions(bioNumber)
        print(spEquivalence)
        useID = False
        # reactionDefinitions = 'reactionDefinitions/reactionDefinition9.json'
        outputFile = "complex/output" + str(bioNumber) + ".bngl"
        analyzeFile(
            "XMLExamples/curated/BIOMD%010i.xml" % bioNumber,
            reactionDefinitions,
            useID,
            naming,
            outputFile,
            speciesEquivalence=spEquivalence,
            atomize=True,
            bioGrid=True,
        )

        if len(logMess.log) > 0:
            with open(outputFile + ".log", "w") as f:
                for element in logMess.log:
                    f.write(element + "\n")


def detectCustomDefinitions(bioNumber):
    """
    returns a speciesDefinition<bioNumber>.json fileName if it exist
    for the current bioModels. None otherwise
    """
    directory = "reactionDefinitions"
    onlyfiles = [f for f in listdir("./" + directory)]
    if "speciesEquivalence{0}.json".format(bioNumber) in onlyfiles:
        return "{0}/speciesEquivalence{1}.json".format(directory, bioNumber)
    return None


import pyparsing


def main():
    jsonFiles = [f for f in listdir("./reactionDefinitions") if f[-4:-1] == "jso"]
    jsonFiles.sort()
    parser = OptionParser()
    rulesLength = []
    evaluation = []
    evaluation2 = []
    compartmentLength = []
    parser.add_option(
        "-i",
        "--input",
        dest="input",
        default="XMLExamples/curated/BIOMD0000000272.xml",
        type="string",
        help="The input SBML file in xml format. Default = 'input.xml'",
        metavar="FILE",
    )
    parser.add_option(
        "-o",
        "--output",
        dest="output",
        default="output.bngl",
        type="string",
        help="the output file where we will store our matrix. Default = output.bngl",
        metavar="FILE",
    )

    (options, _) = parser.parse_args()
    # 144
    rdfArray = []
    # classificationArray = []
    # 18, 32, 87, 88, 91, 109, 253, 255, 268, 338, 330
    # normal:51, 353
    # cycles 18, 108, 109, 255, 268, 392
    import progressbar

    progress = progressbar.ProgressBar()
    sbmlFiles = getFiles("XMLExamples/curated", "xml")
    for bioIdx in progress(range(len(sbmlFiles))):
        bioNumber = sbmlFiles[bioIdx]

        # if bioNumber in [81, 151, 175, 205, 212, 223, 235, 255, 326, 328, 347, 370, 404, 428, 430, 431, 443, 444, 452, 453, 465, 474]:
        #    continue
        # bioNumber = 175
        logMess.log = []
        logMess.counter = -1
        # reactionDefinitions, useID, naming = selectReactionDefinitions('BIOMD%010i.xml' %bioNumber)
        # print(reactionDefinitions, useID)
        # reactionDefinitions = 'reactionDefinitions/reactionDefinition7.json'
        # spEquivalence = 'reactionDefinitions/speciesEquivalence19.json'
        # spEquivalence = naming
        # reactionDefinitions = 'reactionDefinitions/reactionDefinition8.json'
        # rlength, reval, reval2, clength, rdf = analyzeFile('XMLExamples/curated/BIOMD%010i.xml' % bioNumber,
        #                                                  reactionDefinitions, False, 'complex/output' + str(bioNumber) + '.bngl',
        #                                                    speciesEquivalence=spEquivalence, atomize=True)
        try:
            fileName = bioNumber.split("/")[-1]
            rlength = reval = reval2 = slength = None
            analysisResults = analyzeFile(
                bioNumber,
                resource_path("config/reactionDefinitions.json"),
                False,
                resource_path("config/namingConventions.json"),
                #'/dev/null',
                "complex2/" + fileName + ".bngl",
                speciesEquivalence=None,
                atomize=True,
                bioGrid=False,
            )
            # rlength, slength, reval, reval2, clength, rdf, _, _ = analysisResults
            # print('++++', bioNumber, rlength, reval, reval2, clength)

        except KeyError:
            print("keyErrorerror--------------", bioNumber)
            continue
        except OverflowError:
            print("overFlowerror--------------", bioNumber)
            continue
        except ValueError:
            print("valueError", bioNumber)
        except pyparsing.ParseException:
            print("pyparsing", bioNumber)
        finally:
            if analysisResults.rlength != None:
                rulesLength.append(
                    {
                        "index": bioNumber,
                        "nreactions": analysisResults.rlength,
                        "atomization": analysisResults.reval,
                        "compression": analysisResults.reval2,
                        "nspecies": analysisResults.slength,
                    }
                )
                compartmentLength.append(analysisResults.clength)
                rdfArray.append(getAnnotationsDict(analysisResults.rdf))

            else:
                rulesLength.append([bioNumber, -1, 0, 0])
                compartmentLength.append(0)
                rdfArray.append({})

            # classificationArray.append({})
    # print(evaluation)
    # print(evaluation2)
    # sortedCurated = [i for i in enumerate(evaluation), key=lambda x:x[1]]
    print([(idx + 1, x) for idx, x in enumerate(rulesLength) if x > 50])
    with open("sortedD.dump", "wb") as f:
        pickle.dump(rulesLength, f)
    with open("annotations.dump", "wb") as f:
        pickle.dump(rdfArray, f)
    # with open('classificationDict.dump', 'wb') as f:
    #    pickle.dump(classificationArray, f)
    """
    plt.hist(rulesLength, bins=[10, 30, 50, 70, 90, 110, 140, 180, 250, 400])
    plt.xlabel('Number of reactions', fontsize=18)
    plt.savefig('lengthDistro.png')
    plt.clf()
    plt.hist(evaluation, bins=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7,
                                0.8, 0.9, 1.0])
    plt.xlabel('Atomization Degree', fontsize=18)
    plt.savefig('ruleifyDistro.png')
    plt.clf()
    plt.hist(evaluation2, bins=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7,
                                0.8, 0.9, 1.0])
    plt.xlabel('Atomization Degree', fontsize=18)
    plt.savefig('ruleifyDistro2.png')
    plt.clf()
    ev = []
    idx = 1
    for x, y, z in zip(rulesLength, evaluation, compartmentLength):

        if idx in [18, 51, 353, 108, 109, 255, 268, 392]:
            idx+=1

        if x < 15 and y > 0.7 and z>1:
            print('---', idx, x, y)
        idx+=1
    #plt.hist(ev, bins =[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
    #plt.xlabel('Atomization Degree', fontsize=18)
    #plt.savefig('ruleifyDistro3.png')
    """


def main2():
    with open("../XMLExamples/curated/BIOMD0000000163.xml", "r") as f:
        st = f.read()
        import StringIO

        stringBuffer = StringIO.StringIO()
        jsonPointer = "reactionDefinitions/speciesEquivalence163.json"
        readFromString(
            st,
            resource_path("config/reactionDefinitions.json"),
            False,
            jsonPointer,
            True,
            stringBuffer,
        )
        print(stringBuffer.getvalue())


def isActivated(statusVector):
    if statusVector[0] != "" or statusVector[1] not in ["", "0"]:
        return True
    return False


def flatStatusVector(statusVector):
    if statusVector[0] != "":
        return "!"
    return statusVector[1]


"""
def xorBox(status1, status2):
    return not(status1 & status2)

def orBox(status1, status2):
    return (status1, status2)

def totalEnumerations(pairList):
    xCoordinate = set()
    yCoordinate = set()
    for element in pairList:
        xCoordinate.add(element[0])
        yCoordinate.add(element[1])
    xCoordinate = list(xCoordinate)
    yCoordinate = list(yCoordinate)
    matrix = np.zeros((len(xCoordinate), len(yCoordinate)))
    for element in pairList:
        matrix[xCoordinate.index(element[0])][yCoordinate.index(element[1])] = 1
    return np.all(np.all(matrix))
"""


def getRelationshipDegree(
    componentPair, statusQueryFunction, comparisonFunction, finalComparison
):
    componentPairRelationshipDict = {}
    for pair in componentPair:
        stats = []
        for state in componentPair[pair]:
            status1 = statusQueryFunction(state[0])
            status2 = statusQueryFunction(state[1])
            comparison = comparisonFunction(status1, status2)
            stats.append(comparison)
        if finalComparison(stats):
            print(pair, componentPair[pair])
        componentPairRelationshipDict[pair] = finalComparison(stats)
    return componentPairRelationshipDict


"""
def createPlot(labelDict):
    #f, ax = plt.subplots(int(math.ceil(len(labelDict)/4)), 4)
    for idx, element in enumerate(labelDict):
        plt.cla()
        tmp = list(set([y for x in labelDict[element] for y in x]))
        xaxis = [tmp.index(x[0]) for x in labelDict[element] if  labelDict[element][x]== True]
        yaxis = [tmp.index(x[1]) for x in labelDict[element] if labelDict[element][x] == True]
        #6print(tmp, xaxis, yaxis)
        plt.scatter(xaxis, yaxis)
        plt.xticks(range(len(tmp)), tmp)
        plt.yticks(range(len(tmp)), tmp)
        plt.title(element)
        #ax[math.floor(idx/4)][idx%4].scatter(xaxis, yaxis)
        #ax[math.floor(idx/4)][idx%4].xticks(range(len(tmp)), tmp)
        #ax[math.floor(idx/4)][idx%4].yticks(range(len(tmp)), tmp)
        #ax[math.floor(idx/4)][idx%4].title(element)
        plt.savefig('{0}.png'.format(element))
        print('{0}.png'.format(element))
"""
"""
def statFiles():

    for bioNumber in [19]:
        reactionDefinitions, useID = selectReactionDefinitions('BIOMD%010i.xml' %bioNumber)
        #speciesEquivalence = None
        speciesEquivalence = 'reactionDefinitions/speciesEquivalence19.json'

        componentPairs =  extractCompartmentStatistics('XMLExamples/curated/BIOMD%010i.xml' % bioNumber, useID, reactionDefinitions, speciesEquivalence)
        #analyze the relationship degree betweeen the components of each molecule

        #in this case we are analyzing for orBoxes, or components
        #that completely exclude each other
        xorBoxDict = {}
        orBoxDict = {}
        for molecule in componentPairs:
            xorBoxDict[molecule] = getRelationshipDegree(componentPairs[molecule], isActivated, xorBox, all)
            #print('----------------------', molecule, '---------'            )
            orBoxDict[molecule] =  getRelationshipDegree(componentPairs[molecule], flatStatusVector, orBox, totalEnumerations)

        #createPlot(orBoxDict)
        box = []
        box.append(xorBoxDict)
        #box.append(orBoxDict)
        with open('orBox{0}.dump'.format(bioNumber), 'wb') as f:
            pickle.dump(box, f)
"""


def processDir(directory, atomize=True):
    from os import listdir
    from os.path import isfile, join

    resultDir = {}
    xmlFiles = [
        f
        for f in listdir("./" + directory)
        if isfile(join("./" + directory, f)) and f.endswith("xml")
    ]
    blackList = [
        175,
        205,
        212,
        223,
        235,
        255,
        328,
        370,
        428,
        430,
        431,
        443,
        444,
        452,
        453,
        465,
    ]

    for xml in xmlFiles:
        # try:
        if (
            xml not in ["MODEL1310110034.xml"]
            and len([x for x in blackList if str(x) in xml]) == 0
        ):
            print(xml)
            try:
                analysisResults = analyzeFile(
                    directory + xml,
                    "reactionDefinitions/reactionDefinition7.json",
                    False,
                    resource_path("config/namingConventions.json"),
                    "/dev/null",
                    speciesEquivalence=None,
                    atomize=True,
                    bioGrid=False,
                )
                resultDir[xml] = [
                    analysisResults.rlength,
                    analysisResults.reval,
                    analysisResults.reval2,
                ]
            except:
                resultDir[xml] = [-1, 0, 0]
    with open("evalResults.dump", "wb") as f:
        pickle.dump(resultDir, f)
        # except:
        # continue'


def processFile3(
    fileName, customDefinitions=None, atomize=True, bioGrid=False, output=None
):
    """
    processes a file. derp.
    """
    logMess.log = []
    logMess.counter = -1
    reactionDefinitions = resource_path("config/reactionDefinitions.json")
    spEquivalence = customDefinitions
    namingConventions = resource_path("config/namingConventions.json")
    # spEquivalence = None
    useID = False
    # reactionDefinitions = 'reactionDefinitions/reactionDefinition9.json'
    # rlength = -1
    # reval = -1
    # reval2 = -1
    if output:
        outputFile = output
    else:
        outputFile = "{0}.bngl".format(fileName)
    analysisResults = analyzeFile(
        fileName,
        reactionDefinitions,
        useID,
        namingConventions,
        outputFile,
        speciesEquivalence=spEquivalence,
        atomize=atomize,
        bioGrid=bioGrid,
    )

    if len(logMess.log) > 0:
        with open(fileName + ".log", "w") as f:
            for element in logMess.log:
                f.write(element + "\n")
    return analysisResults.rlength, analysisResults.reval, analysisResults.reval2


def listFiles(minReactions, directory):
    """
    List of SBML files that meet a given condition
    """
    from os import listdir
    from os.path import isfile, join

    xmlFiles = [
        f
        for f in listdir("./" + directory)
        if isfile(join("./" + directory, f)) and "xml" in f
    ]
    outputList = []
    for xml in xmlFiles:
        print(
            ".",
        )
        reader = libsbml.SBMLReader()
        document = reader.readSBMLFromFile(directory + xml)
        model = document.getModel()
        if model == None:
            continue
        if len(model.getListOfReactions()) > minReactions:
            outputList.append(xml)
    print(len(outputList))


if __name__ == "__main__":
    main2()



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 27 14:45:24 2012

@author: proto
"""

"""
this method classifies reactants according to the rdf information, and gives
us information on which reactants are the same, and how do they differ
(compartment etc)
"""
try:
    from sbml2bngl import SBML2BNGL
    import libsbml
except ModuleNotFoundError:
    import sys

    sys.path.append("..")
    from sbml2bngl import SBML2BNGL
    import libsbml

import collections


def getAnnotations(parser, stringKey=None):
    annotation = parser.getSpeciesAnnotation()
    annotationDictionary = collections.defaultdict(set)
    for key, value in annotation.items():
        annotationList = []
        if annotation[key] != None:
            for element in annotation[key]:
                for index in range(0, element.getNumAttributes()):
                    if not stringKey or stringKey in element.getValue(index):
                        annotationList.append(element.getValue(index))
                if annotationList == []:
                    continue
                annotationDictionary[key].add(tuple(annotationList))
                # annotationDictionary[frozenset(annotationList)].sort(lambda x,y: cmp(len(x), len(y)))
    return annotationDictionary


def getSpeciesAnnotationStructure(parser):
    model = parser.model
    for species in model.getListOfSpecies():
        name = species.getName()
        speciesId = species.getId()
        annotation = species.getAnnotation()
        lista = libsbml.CVTermList()
        libsbml.RDFAnnotationParser.parseRDFAnnotation(annotation, lista)
        for idx in range(0, lista.getSize()):
            for idx2 in range(0, lista.get(idx).getResources().getLength()):
                resource = lista.get(idx).getResources().getValue(idx2)
                qualifierType = lista.get(idx).getQualifierType()
                qualifierDescription = (
                    bioqual[lista.get(idx).getBiologicalQualifierType()]
                    if qualifierType
                    else modqual[lista.get(idx).getModelQualifierType()]
                )
                # resource = resolveAnnotation(resource)


def getEquivalence(species, rdf_database):
    """
    *species* is the species whose equivalence we will go and search
    This method will search through the RDF database and look if param 'species'
    is equal to any other element in the species database
    """

    for element in rdf_database:
        if species in rdf_database[element]:
            if rdf_database[element].index(species) == 0:
                return []
            # return [x for x in rdf_database[element] if x != species]

            # well only return the first one by default
            return [rdf_database[element][0]]
    return []


if __name__ == "__main__":
    reader = libsbml.SBMLReader()
    # BIOMD0000000272
    document = reader.readSBMLFromFile("XMLExamples/curated/BIOMD0000000219.xml")
    # document = reader.readSBMLFromFile('XMLExamples/simple4.xml')
    model = document.getModel()
    parser = SBML2BNGL(model)
    annotationDictionary = getAnnotations(parser)
    print(annotationDictionary)
    # print(getEquivalence('SAv_EpoR',annotationDictionary))
    # print(annotation)
    # print(rules    )



import copy

from bionetgen.main import BioNetGen
from tempfile import TemporaryFile
from .bngparser import BNGParser
from .blocks import (
    ActionBlock,
    CompartmentBlock,
    FunctionBlock,
    MoleculeTypeBlock,
    ObservableBlock,
    ParameterBlock,
    RuleBlock,
    SpeciesBlock,
    EnergyPatternBlock,
)


# This allows access to the CLIs config setup
app = BioNetGen()
app.setup()
conf = app.config["bionetgen"]
def_bng_path = conf["bngpath"]

###### CORE OBJECT AND PARSING FRONT-END ######
class bngmodel:
    """
    Main model object and entry point for model API. The goal of this
    object is to generate and read the BNGXML of a given BNGL model
    and give the user a pythonic interface to the resulting model object.

    Usage: bngmodel(bng_model)
           bngmodel(bng_model, BNGPATH)

    Attributes
    ----------
    active_blocks : list[str]
        a list of the blocks that have been parsed in the model
    bngparser : BNGParser
        BNGParser object that's responsible for .bngl file reading and model setup
    model_name : str
        name of the model, generally set from the given BNGL file
    recompile : bool
        a tag to keep track if any changes have been made to the model
        via the XML API by the user
    changes : dict
        a list of changes the user have made to the model

    Methods
    -------
    reset_compilation_tags()
        resets compilation tags of each block to keep track of any changes the user
        makes to the model via the API
    add_action(action_type, action_args)
        adds the action of action_type with arguments given by the optional keyword
        argument action_args which is a list of lists where each element
        is of the form [ArgumentName, ArgumentValue]
    write_model(model_name)
        write the model in BNGL format to the path given
    setup_simulator(sim_type)
        sets up a simulator in bngmodel.simulator where the only current supported
        type of simulator is libRR for libRoadRunner simulator.
    """

    def __init__(self, bngl_model, BNGPATH=def_bng_path):
        self.active_blocks = []
        # We want blocks to be printed in the same order every time
        self.block_order = [
            "parameters",
            "compartments",
            "molecule_types",
            "species",
            "energy_patterns",
            "observables",
            "functions",
            "rules",
            "actions",
        ]
        self.model_name = ""
        self.bngparser = BNGParser(bngl_model)
        self.bngparser.parse_model(self)
        for block in self.block_order:
            if block not in self.active_blocks:
                self.add_empty_block(block)
        # Check to see if there are no active blocks
        # If not, model is most likely not in BNGL format
        if not self.active_blocks:
            print(
                "WARNING: No active blocks. Please ensure model is in proper BNGL or BNG-XML format"
            )

    @property
    def recompile(self):
        recompile = False
        for block in self.active_blocks:
            recompile = recompile or getattr(self, block)._recompile
        return recompile

    @property
    def changes(self):
        changes = {}
        for block in self.active_blocks:
            changes[block] = getattr(self, block)._changes
        return changes

    def __str__(self):
        """
        write the model to str
        """
        model_str = ""
        # gotta check for "before model" type actions
        if hasattr(self, "actions"):
            ablock = getattr(self, "actions")
            if len(ablock.before_model) > 0:
                for baction in ablock.before_model:
                    model_str += str(baction) + "\n"
        model_str += "begin model\n"
        for block in self.block_order:
            # ensure we didn't get new items into a
            # previously inactive block, if we did
            # add them to the active blocks
            if hasattr(self, block):
                if len(getattr(self, block)) > 0:
                    if getattr(self, block).name not in self.active_blocks:
                        self.active_blocks.append(block)
                # if we removed items from a block and
                # it's now empty, we want to remove it
                # from the active blocks
                elif len(getattr(self, block)) == 0 and block in self.active_blocks:
                    self.active_blocks.remove(block)
            # print only the active blocks
            if block in self.active_blocks:
                if block != "actions" and len(getattr(self, block)) > 0:
                    model_str += str(getattr(self, block))
        model_str += "\nend model\n\n"
        if "actions" in self.active_blocks:
            model_str += str(self.actions)
        return model_str

    def __repr__(self):
        return self.model_name

    def __iter__(self):
        active_ordered_blocks = [
            getattr(self, i) for i in self.block_order if i in self.active_blocks
        ]
        return active_ordered_blocks.__iter__()

    def add_block(self, block):
        bname = block.name.replace(" ", "_")
        # TODO: fix this exception
        if bname == "reaction_rules":
            bname = "rules"
        block_adder = getattr(self, "add_{}_block".format(bname))
        block_adder(block)

    def add_empty_block(self, block_name):
        bname = block_name.replace(" ", "_")
        # TODO: fix this exception
        if bname == "reaction_rules":
            bname = "rules"
        block_adder = getattr(self, "add_{}_block".format(bname))
        block_adder()

    def add_parameters_block(self, block=None):
        if block is not None:
            assert isinstance(block, ParameterBlock)
            self.parameters = block
            if "parameters" not in self.active_blocks:
                self.active_blocks.append("parameters")
        else:
            self.parameters = ParameterBlock()

    def add_compartments_block(self, block=None):
        if block is not None:
            assert isinstance(block, CompartmentBlock)
            self.compartments = block
            if "compartments" not in self.active_blocks:
                self.active_blocks.append("compartments")
        else:
            self.compartments = CompartmentBlock()

    def add_molecule_types_block(self, block=None):
        if block is not None:
            assert isinstance(block, MoleculeTypeBlock)
            self.molecule_types = block
            if "molecule_types" not in self.active_blocks:
                self.active_blocks.append("molecule_types")
        else:
            self.molecule_types = MoleculeTypeBlock()

    def add_species_block(self, block=None):
        if block is not None:
            assert isinstance(block, SpeciesBlock)
            self.species = block
            if "species" not in self.active_blocks:
                self.active_blocks.append("species")
        else:
            self.species = SpeciesBlock()

    def add_observables_block(self, block=None):
        if block is not None:
            assert isinstance(block, ObservableBlock)
            self.observables = block
            if "observables" not in self.active_blocks:
                self.active_blocks.append("observables")
        else:
            self.observables = ObservableBlock()

    def add_functions_block(self, block=None):
        if block is not None:
            assert isinstance(block, FunctionBlock)
            self.functions = block
            if "functions" not in self.active_blocks:
                self.active_blocks.append("functions")
        else:
            self.functions = FunctionBlock()

    def add_rules_block(self, block=None):
        if block is not None:
            assert isinstance(block, RuleBlock)
            self.rules = block
            if "rules" not in self.active_blocks:
                self.active_blocks.append("rules")
        else:
            self.rules = RuleBlock()

    def add_energy_patterns_block(self, block=None):
        if block is not None:
            assert isinstance(block, EnergyPatternBlock)
            self.energy_patterns = block
            if "energy_patterns" not in self.active_blocks:
                self.active_blocks.append("energy_patterns")
        else:
            self.energy_patterns = EnergyPatternBlock()

    def add_actions_block(self, block=None):
        if block is not None:
            assert isinstance(block, ActionBlock)
            self.actions = block
            if "actions" not in self.active_blocks:
                self.active_blocks.append("actions")
        else:
            self.actions = ActionBlock()

    def reset_compilation_tags(self):
        for block in self.active_blocks:
            getattr(self, block).reset_compilation_tags()

    def add_action(self, action_type, action_args=[]):
        # add actions block and to active list
        if not hasattr(self, "actions"):
            self.actions = ActionBlock()
            if "actions" not in self.active_blocks:
                self.active_blocks.append("actions")
        self.actions.add_action(action_type, action_args)

    def write_model(self, file_name):
        """
        write the model to file
        """
        model_str = ""
        for block in self.active_blocks:
            model_str += str(getattr(self, block))
        with open(file_name, "w") as f:
            f.write(model_str)

    def setup_simulator(self, sim_type="libRR"):
        """
        Sets up a simulator attribute that is a generic front-end
        to all other simulators. At the moment only libroadrunner
        is supported
        """
        if sim_type == "libRR":
            # we need to add writeSBML action for now
            curr_actions = copy.deepcopy(self.actions)
            self.actions.clear_actions()
            self.add_action("generate_network", {"overwrite": 1})
            self.add_action("writeSBML", {})
            # temporary file
            with TemporaryFile(mode="w+") as tpath:
                # write the sbml
                if not (
                    self.bngparser.bngfile.write_xml(
                        tpath, xml_type="sbml", bngl_str=str(self)
                    )
                ):
                    raise ValueError("SBML couldn't be generated for libRR simulator")
                # TODO: Only clear the writeSBML action
                # by adding a mechanism to do so
                self.actions.clear_actions()
                # get the simulator
                import bionetgen as bng

                self.simulator = bng.sim_getter(
                    model_str=tpath.read(), sim_type=sim_type
                )
                # let's deal with observables here
                selections = ["time"] + [obs for obs in self.observables]
                self.simulator.simulator.timeCourseSelections = selections
            self.actions = curr_actions
        else:
            print(
                'Sim type {} is not recognized, only libroadrunner \
                   is supported currently by passing "libRR" to \
                   sim_type keyword argument'.format(
                    sim_type
                )
            )
            return None
        # for now we return the underlying simulator
        return self.simulator.simulator


###### CORE OBJECT AND PARSING FRONT-END ######



import bioservices
import pyparsing as pyp

goGrammar = (
    pyp.Suppress(pyp.Literal("<name>"))
    + pyp.Word(pyp.alphanums + " -_/")
    + pyp.Suppress(pyp.Literal("</name>"))
)

from contextlib import contextmanager
import sys, os


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = devnull
        sys.stderr = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


def resolveAnnotation(annotation):
    with suppress_stdout():
        return resolveAnnotationHelper(annotation)


def resolveAnnotationHelper(annotation):
    if not hasattr(resolveAnnotation, "db"):
        resolveAnnotation.db = {}
        resolveAnnotation.ch = bioservices.ChEBI(verbose=False)
        resolveAnnotation.uni = bioservices.UniProt(verbose=False)
        resolveAnnotation.k = bioservices.kegg.KEGG(verbose=False)
        resolveAnnotation.qg = bioservices.QuickGO(verbose=False)
        resolveAnnotation.t = bioservices.Taxon()
        resolveAnnotation.db[
            "http://identifiers.org/uniprot/P62988"
        ] = "http://identifiers.org/uniprot/P62988"
        resolveAnnotation.db[
            "http://identifiers.org/uniprot/P06842"
        ] = "http://identifiers.org/uniprot/P06842"
        resolveAnnotation.db[
            "http://identifiers.org/uniprot/P07006"
        ] = "http://identifiers.org/uniprot/P06842"

    if annotation in resolveAnnotation.db:
        return annotation, resolveAnnotation.db[annotation]

    tAnnotation = annotation.replace("%3A", ":")
    tAnnotation = annotation.split("/")[-1]
    # tAnnotation = re.search(':([^:]+:[^:]+$)',tAnnotation).group(1)
    try:
        if "obo.go" in annotation or "/go/GO" in annotation:

            res = resolveAnnotation.qg.Term(tAnnotation)
            finalArray = []
            if type(res) not in [int]:
                res = bioservices.Service("name").easyXML(res)
                tmp = res.findAll("name")

                for x in tmp:
                    try:
                        tagString = str(goGrammar.parseString(str(x))[0])
                        if tagString not in ["Systematic synonym"]:
                            finalArray.append(str(goGrammar.parseString(str(x))[0]))
                    except pyp.ParseBaseException:
                        continue
            if len(finalArray) > 0:
                resolveAnnotation.db[annotation] = finalArray[0]
            else:
                resolveAnnotation.db[annotation] = ""
            finalAnnotation = resolveAnnotation.db[annotation]

        elif "kegg" in annotation:

            data = resolveAnnotation.k.get(tAnnotation)
            dict_data = resolveAnnotation.k.parse(data)
            if type(dict_data) == int:
                resolveAnnotation.db[annotation] = ""
            else:
                resolveAnnotation.db[annotation] = dict_data["name"]
            finalAnnotation = resolveAnnotation.db[annotation]

        elif "uniprot" in annotation:
            identifier = annotation.split("/")[-1]
            result = resolveAnnotation.uni.quick_search(identifier)
            if identifier in result:
                resolveAnnotation.db[annotation] = result[identifier][
                    "Protein names"
                ].split("(")[0]
            else:
                finalAnnotation = ""
            finalAnnotation = resolveAnnotation.db[annotation]

        elif "chebi" in annotation:
            tmp = annotation.split("/")[-1]

            entry = resolveAnnotation.ch.getLiteEntity(tmp)
            finalAnnotation = ""
            for element in entry:
                resolveAnnotation.db[annotation] = str(element["chebiAsciiName"])
                finalAnnotation = resolveAnnotation.db[annotation]
        elif (
            "cco" in annotation
            or "pirsf" in annotation
            or "pubchem" in annotation
            or "omim" in annotation
        ):
            finalAnnotation = ""
        elif "taxonomy" in annotation:
            # uniprot stuff for taxonomy
            result = resolveAnnotation.t.search_by_taxon(tAnnotation)
            resolveAnnotation.db[annotation] = result["Scientific Name"]
            finalAnnotation = resolveAnnotation.db[annotation]
            """
            url = 'http://www.uniprot.org/taxonomy/'
            params = {
            'from':'ACC',
            'to':'P_REFSEQ_AC',
            'format':'tab',
            'query':'P13368 P20806 Q9UM73 P97793 Q17192'
            }
            
            data = urllib.urlencode(params)
            request = urllib2.Request(url, data)
            contact = "" # Please set your email address here to help us debug in case of problems.
            request.add_header('User-Agent', 'Python contact')
            response = urllib2.urlopen(request)
            page = response.read(200000)
            """
        else:
            return annotation, ""
            # assert(False)
            finalAnnotation = ""
    except (IOError, KeyError) as e:
        return annotation, ""
    return annotation, finalAnnotation


if __name__ == "__main__":
    print resolveAnnotation("http://identifiers.org/taxonomy/10116")
    # print resolveAnnotation('http://identifiers.org/uniprot/P01133')



import namingDatabase


def createNamingMatrix(fileList):
    pass


if __name__ == "__main__":
    db = namingDatabase.NamingDatabase("namingConventions.db")
    fileList = ["BIOMD0000000205.xml", "BIOMD0000000048.xml", "BIOMD0000000019.xml"]
    createNamingMatrix(fileList)



# -*- coding: utf-8 -*-
"""
Created on Wed Aug  7 21:04:16 2013

@author: proto
"""

import utils.consoleCommands as console
from utils.smallStructures import readFromString
import utils.extractAtomic as extractAtomic
import utils.readBNGXML as readBNGXML
from collections import Counter
import numpy as np


def readFile(fileName):
    f = open(fileName)
    r = f.readlines()
    return r


def getPatternDict(patternList):
    """
    goes through a list of atomic patterns and orders them in a dictionary according to the molecule/
    component pair they contain. This means that molecules with the same molecule/component pair but
    different number of bonds or state configuration will be grouped together
    """
    patternDict = {}
    extendedList = list(patternList.elements())
    stExtendList = [readFromString(x) for x in extendedList]

    for species in stExtendList:
        for molecule in species.molecules:
            if molecule.str3() not in patternDict:
                patternDict[molecule.str3()] = []
            patternDict[molecule.str3()].append(str(species))

    return patternDict


def matchElements(moleculeSet1, moleculeSet2):
    """
    create a distance matrix set1 vs set2 and choose elements from different rows/columns that minimize
    distance
    """

    distanceMatrix = [
        [0 for x in range(0, len(moleculeSet2))] for x in range(0, len(moleculeSet1))
    ]

    for idx, reference in enumerate(moleculeSet1):
        for idx2, comparison in enumerate(moleculeSet2):
            distanceMatrix[idx][idx2] = reference[1].distance(comparison[1])
    elements = []
    for row in distanceMatrix:
        while row.index(min(row)) in elements:
            row[row.index(min(row))] = 100000
        elements.append(row.index(min(row)))
    finalArray = []
    for idx, element in enumerate(elements):
        finalArray.append((moleculeSet1[idx], moleculeSet2[element]))
    return finalArray


def getMetaElement(matchedArray):
    for element in matchedArray:
        element[0][1].compare(element[1][1])


def createMetaRule(ruleSet, differences):
    """
    Creates a metaRule from an array 'ruleSet' of rules. The differences parameter contains a dictionary
    elaborating on how the rules are different
    """
    moleculeDict = []
    for ruleDescription in ruleSet:
        # todo:i have to find the way to group together equivalent
        # molecules from different rules and find the metarule
        molList = {}
        for reactant in ruleDescription[0].reactants:
            for key in differences:
                for molecule in reactant.molecules:
                    if molecule.name + "(" in key:
                        for component in molecule.components:
                            if "(" + component.name + ")" in key:
                                # print molecule.name, component.name, key
                                if key not in molList:
                                    molList[key] = []
                                molList[key].append([reactant, molecule, component])
        moleculeDict.append(molList)
        for reactant in ruleDescription[0].products:
            for key in differences:
                for molecule in reactant.molecules:
                    if molecule.name + "(" in key:
                        for component in molecule.components:
                            if "(" + component.name + ")" in key:
                                # print molecule.name, component.name, key
                                if key not in molList:
                                    molList[key] = []
                                molList[key].append([reactant, molecule, component])
        moleculeDict.append(molList)

    metaRule = moleculeDict[0]
    matchedArray = {}
    for idx in range(1, len(moleculeDict)):

        for element in metaRule:
            if element in moleculeDict[idx]:
                matchedArray = matchElements(
                    metaRule[element], moleculeDict[idx][element]
                )
                getMetaElement(matchedArray)
            # print metaRule[element], moleculeDict[idx][element]


def groupByReactionCenter(transformationCenter):
    """
    returns:  A Dictionary with 'reactionCenter' keys.
    """
    centerDict = {}
    # extract rules with teh same reaction center
    for idx, rule in enumerate(transformationCenter):
        # print rule
        element = [tuple(x.elements()) for x in rule if len(x) > 0]
        # print element
        # for element in rule:
        #    if element != set([]):
        if tuple(element) not in centerDict:
            centerDict[tuple(element)] = set([])
        centerDict[tuple(element)].add(idx)
    return centerDict


def groupByReactionCenterAndRateAndActions(rules, centerDict):
    """
    group those reactions with the same reaction center on the first level
    and reaction rate/ actions on the second level
    """
    redundantDict = {}
    for center in centerDict:
        tmpDict = {}
        for rnum in centerDict[center]:
            actionString = rules[rnum][0].actions
            key = [x.action for x in actionString]
            key.sort()
            # append the rate to the tuple to create a key
            key.append(rules[rnum][-2])
            if tuple(key) not in tmpDict:
                tmpDict[tuple(key)] = []
            tmpDict[tuple(key)].append(rnum)
        for rate in tmpDict:
            if len(tmpDict[rate]) >= 1:
                if center not in redundantDict:
                    redundantDict[center] = {}
                redundantDict[center][rate] = tmpDict[rate]
    return redundantDict


def groupByReactionCenterAndRateAndActions2(rules, centerDict):
    """
    group those reactions with the same reaction center on the first level
    and reaction rate/ actions on the second level
    """
    redundantDict = {}
    for center in centerDict:
        tmpDict = {}
        for rnum in centerDict[center]:
            actionString = rules[rnum][0].actions
            key = [x.action for x in actionString]
            key.sort()
            # append the rate to the tuple to create a key
            key.append(rules[rnum][-2])
            if tuple(key) not in tmpDict:
                tmpDict[tuple(key)] = []
            tmpDict[tuple(key)].append(rnum)
        for rate in tmpDict:
            if len(tmpDict[rate]) >= 1:
                center2 = list(center)
                center2.extend(rate[:-1])
                center2 = tuple(center2)
                if center2 not in redundantDict:
                    redundantDict[center2] = {}

                redundantDict[center2][rate[-1]] = tmpDict[rate]
    return redundantDict


def obtainDifferences(redundantDict, transformationContext):
    """
    analize what is different between similar rules (same rc, rate and actions) throw unions
    and intersections of           their context
    """
    redundantListDict = {}
    for center in redundantDict:
        for rate in redundantDict[center]:
            tmp = []
            tmp = [transformationContext[x] for x in redundantDict[center][rate]]
            # if center == (('Prot(iMod~U)', ), ('Prot(egfr)', 'EGFR(prot)')):
            #    print '--', tmp

            # pair together by traqnsformation the context of all those rules whose reaction center, rate
            # and transformations are the same
            tmp = zip(*tmp)
            # if center == (('Prot(iMod~U)', ), ('Prot(egfr)', 'EGFR(prot)')):
            #    print '--------------------------------'
            #    print '--', tmp
            # print '@@@', tmp
            intersections = []
            unions = []
            for ttmp in tmp:
                tint = ttmp[0]
                tun = ttmp[0]
                for idx in range(1, len(ttmp)):
                    tint &= ttmp[idx]
                    tun |= ttmp[idx]
                intersections.append(tint)
                unions.append(tun)
            differences = []
            # obtain the union -intersection per transformation
            for x, y in zip(intersections, unions):
                differences.append(y - x)
            # obtain those elements that were found to be constantly different across all intersections
            constantDifferences = differences[0]
            for idx in range(1, len(differences)):
                constantDifferences &= differences[idx]
            if center not in redundantListDict:
                redundantListDict[center] = {}
            redundantListDict[center][rate] = constantDifferences
    return redundantListDict


# XXX: How was this supposed to work. pgv is never imported.
#
# def reactionCenterGraph(species, reactionCenter):
#     total = sum(x[1] for x in reactionCenter)
#     graph = pgv.AGraph(directed=False,concentrate=True)
#     print reactionCenter,
#     for element in species:
#         graph.add_node(element.name, shape='diamond', style='filled')
#         for component in element.components:
#             pass


def extractStatistics():
    number = 151
    console.bngl2xml("complex/output{0}.bngl".format(number))
    species, rules, parameterDict = readBNGXML.parseXML("output{0}.xml".format(number))

    transformationCenter = []
    transformationContext = []

    k = []
    actions = Counter()
    actionSet = Counter()
    for idx, rule in enumerate(rules):
        (
            tatomicArray,
            ttransformationCenter,
            ttransformationContext,
            tproductElements,
            tactionNames,
            tlabelArray,
        ) = extractAtomic.extractTransformations([rule])
        # atomicArray.append(tatomicArray)
        transformationCenter.append(ttransformationCenter)
        transformationContext.append(ttransformationContext)
        k.append(len(rule[0].actions))
        # if len(rule[0].actions) > 3:
        #    print rule[0].reactants
        actions.update([x.action for x in rule[0].actions])
        tmp = [x.action for x in rule[0].actions]
        tmp.sort()
        actionSet.update([tuple(tmp)])

    print("number of species", len(species))
    print("avg number o actions", np.average(k), np.std(k))
    centerDict = groupByReactionCenter(transformationCenter)
    print(
        "singletons",
        len({x: centerDict[x] for x in centerDict if len(centerDict[x]) == 1}),
    )
    tmp = [[tuple(set(x)), len(centerDict[x])] for x in centerDict]
    # reactionCenterGraph(species, tmp)
    # tmp.sort(key=lambda x:x[1], reverse=True)
    print("number of reaction centers", len(centerDict.keys()))
    print("number of rules", len(rules))

    # print 'unique', [x for x in centerDict[element] if len(centerDict[element]) == 1]
    redundantDict = groupByReactionCenterAndRateAndActions(rules, centerDict)
    # print redundantDict
    tmp2 = [
        (
            "$\\tt{{{0}}}$".format(tuple(set(x))),
            tuple(set(y[:-1])),
            y[-1],
            len(redundantDict[x][y]),
        )
        for x in redundantDict
        for y in redundantDict[x]
        if "kr" not in y[-1]
    ]
    tmp2 = set(tmp2)
    tmp2 = list(tmp2)
    tmp2.sort(key=lambda x: x[3], reverse=True)
    tmp2.sort(key=lambda x: x[0], reverse=True)

    tmp2 = [
        "{0} & {1} & {2} & {3}\\\\\n".format(
            element[0], element[1], element[2], element[3]
        )
        for element in tmp2
    ]

    with open("table.tex", "w") as f:
        f.write("\\begin{tabular}{|cccc|}\n")
        f.write("\\hline\n")
        f.write("Reaction Centers & Action & Score\\\\\\hline\n")
        for element in tmp2:
            f.write(element)
        f.write("\\hline\n")
        f.write("\\end{tabular}\n")

    x = [len(centerDict[x]) for x in centerDict]
    print(
        "average number of reactions with same rate and reaction cneter",
        np.average(x),
        np.std(x),
    )
    print("total number of clusters", len(x))
    print(x)


def findNewParameters(parameters, bngParameters):
    """
    finds which parameters were added in the xml definition that were not in the original bngl file
    """
    tmp = [par.strip().split(" ")[0] for par in parameters]
    newPar = []
    for bngp in bngParameters:
        if bngp not in tmp:
            newPar.append("\t {0} {1}\n".format(bngp, bngParameters[bngp]))
    return newPar


def extractRedundantContext(rules, transformationCenter, transformationContext):
    """'"""
    centerDict = groupByReactionCenter(transformationCenter)
    redundantDict = groupByReactionCenterAndRateAndActions2(rules, centerDict)
    # redundantDict['{0}.{1}'.format(element, element2)] = tmpDict[element2]
    redundantListDict = obtainDifferences(redundantDict, transformationContext)
    # todo: remove redundancies from rules
    # group together equivalent patterns
    patternDictList = {}
    for center in redundantListDict:
        for rate in redundantListDict[center]:
            patternDict = getPatternDict(redundantListDict[center][rate])
            if center not in patternDictList:
                patternDictList[center] = {}
            patternDictList[center][rate] = patternDict
    return redundantDict, patternDictList


def main():
    fileName = "complex/output19.bngl"
    console.bngl2xml(fileName)
    species, rules, par = readBNGXML.parseXML("output19.xml")
    # print rules

    transformationCenter = []
    transformationContext = []

    # extract the context of such reactions
    for idx, rule in enumerate(rules):
        (
            tatomicArray,
            ttransformationCenter,
            ttransformationContext,
            tproductElements,
            tactionNames,
            tlabelArray,
        ) = extractAtomic.extractTransformations([rule])
        # atomicArray.append(tatomicArray)
        transformationCenter.append(ttransformationCenter)
        transformationContext.append(ttransformationContext)
    redundantDict, patternDictList = extractRedundantContext(
        rules, transformationCenter, transformationContext
    )
    # print redundantDict
    # construct rule based patterns based on the reaction patterns they have to match and
    # the ones they  have to discriminate

    for center in patternDictList:
        for rate in patternDictList[center]:
            match = patternDictList[center][rate]
            notContext = []
            for cRate in [x for x in patternDictList[center] if x != rate]:
                notContext.append(
                    [transformationContext[x] for x in redundantDict[center][cRate]]
                )
            ruleSet = [rules[x] for x in redundantDict[center][rate]]
            createMetaRule(ruleSet, match)

    newRules = range(0, len(rules))
    for center in redundantDict:
        for context in redundantDict[center]:
            for element in range(1, len(redundantDict[center][context])):
                newRules.remove(redundantDict[center][context][element])

    # for element in newRules:
    #    print str(rules[element][0])

    newRulesArray = []
    for element in newRules:
        newRulesArray.append("{0}\n".format(str(rules[element][0])))
    lines = readFile(fileName)
    startR = lines.index("begin reaction rules\n")
    endR = lines.index("end reaction rules\n")
    startP = lines.index("begin parameters\n")
    endP = lines.index("end parameters\n")

    newPar = findNewParameters(lines[startP + 1 : endP], par)
    newLines = (
        lines[0:endP]
        + newPar
        + lines[endP : startR + 1]
        + newRulesArray
        + lines[endR : len(lines)]
    )

    f = open(fileName + "reduced.bngl", "w")
    f.writelines(newLines)
    """
    for reactionCenter in redundantDict:
        candidateRules = [rules[idx][0] for idx in redundantDict[reactionCenter]]
        print '+++', redundantListDict[reactionCenter]
        print '---', candidateRules
    """


if __name__ == "__main__":
    # extractStatistics()
    main()



import annotationExtractor as annEx
import argparse
import fnmatch
import argparse
import os
import progressbar
import cPickle as pickle
import numpy as np

# import SBMLparser.utils.characterizeAnnotationLog as cal


def defineConsole():
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument(
        "-m1", "--model1", type=str, help="input SBML model1", required=True
    )
    parser.add_argument(
        "-m2", "--model2", type=str, help="input SBML model2", required=True
    )
    # parser.add_argument('-o','--output-file',type=str,help='output SBML file',required=True)
    return parser


def componentAnalysis(directory):
    componentCount = []
    bindingCount = []
    stateCount = []
    modelComponentDict = {}
    with open(os.path.join(directory, "moleculeTypeDataSet.dump"), "rb") as f:
        moleculeTypesArray = pickle.load(f)
    for model in moleculeTypesArray:
        modelComponentCount = [len(x.components) for x in model[0]]

        bindingComponentCount = [
            len([y for y in x.components if len(y.states) == 0]) for x in model[0]
        ]

        modificationComponentCount = [
            sum([max(1, len(y.states)) for y in x.components]) for x in model[0]
        ]

        modelComponentDict[model[-2]] = {
            "bindingComponents": np.mean(bindingComponentCount),
            "stateComponents": np.mean(modificationComponentCount),
            "componentCount": np.mean(modelComponentCount),
        }
        # bindingCount.append(bindingComponentCount)
        # stateCount.append(modificationComponentCount)
        # componentCount.append(modelComponentCount)

    # interestingCount = [index for index, x in enumerate(componentCount) if np.mean(x) >= atomizeThreshold]
    # componentCount = np.array(componentCount)
    # bindingCount = np.array(bindingCount)
    # stateCount = np.array(stateCount)

    return modelComponentDict
    # plt.show()


def getFiles(directory, extension):
    """
    Gets a list of <*.extension> files. include subdirectories and return the absolute
    path. also sorts by size.
    """
    matches = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, "*.{0}".format(extension)):
            matches.append(
                [
                    os.path.join(os.path.abspath(root), filename),
                    os.path.getsize(os.path.join(root, filename)),
                ]
            )

    # sort by size
    # matches.sort(key=lambda filename: filename[1], reverse=False)

    matches = [x[0] for x in matches]

    return matches


def annotationComparison(model1, model2, errorList):
    try:
        annotationExtractor = annEx.AnnotationExtractor(model1)
    except AttributeError:
        # print model1, "doesnt exist"
        return -1
    try:
        modelAnnotations1 = annotationExtractor.getModelAnnotations()
    except AttributeError:
        return -1
    # elementalMolecules = [x for x in annotationExtractor.sct if annotationExtractor.sct[x] == []]
    # complexMolecules1 = [x for x in annotationExtractor.sct if annotationExtractor.sct[x] != []]

    annotationDict1 = annotationExtractor.getAnnotationSystem()

    annotationExtractor = annEx.AnnotationExtractor(model2)
    modelAnnotations2 = annotationExtractor.getModelAnnotations()
    # elementalMolecules = [x for x in annotationExtractor.sct if annotationExtractor.sct[x] == []]
    # complexMolecules2 = [x for x in annotationExtractor.sct if annotationExtractor.sct[x] != [] and x in complexMolecules1]
    annotationDict2 = annotationExtractor.getAnnotationSystem()

    error = 0
    for entry in annotationDict1:
        if entry not in annotationDict2:
            continue
        # for label in ['BQB_HAS_PART','BQB_IS_VERSION_OF','BQB_IS','']
        if not set(
            [x for x in annotationDict2[entry]["BQB_HAS_PART"] if "uniprot" in x]
        ).issubset(
            set([x for x in annotationDict1[entry]["BQB_HAS_PART"] if "uniprot" in x])
        ) and not set(
            [x for x in annotationDict2[entry]["BQB_HAS_PART"] if "uniprot" in x]
        ).issubset(
            set(
                [x for x in annotationDict1[entry]["BQB_HAS_VERSION"] if "uniprot" in x]
            )
        ):
            error += 1

        if not set(
            [x for x in annotationDict2[entry]["BQB_HAS_VERSION"] if "uniprot" in x]
        ).issubset(
            set(
                [x for x in annotationDict1[entry]["BQB_HAS_VERSION"] if "uniprot" in x]
            )
        ) and not set(
            [x for x in annotationDict2[entry]["BQB_HAS_VERSION"] if "uniprot" in x]
        ).issubset(
            set([x for x in annotationDict1[entry]["BQB_HAS_PART"] if "uniprot" in x])
        ):
            error += 1

    if error > 0:
        errorList.extend([model1.split("/")[1]])
        return 0
    return 1


def annotationFileComparison(model1, model2):
    try:
        annotationExtractor = annEx.AnnotationExtractor(model1)
    except AttributeError:
        # print model1, "doesnt exist"
        return -1
    modelAnnotations1 = annotationExtractor.getModelAnnotations()
    # elementalMolecules = [x for x in annotationExtractor.sct if annotationExtractor.sct[x] == []]

    annotationDict1 = annotationExtractor.getAnnotationSystem()

    annotationExtractor = annEx.AnnotationExtractor(model2)
    modelAnnotations2 = annotationExtractor.getModelAnnotations()
    # elementalMolecules = [x for x in annotationExtractor.sct if annotationExtractor.sct[x] == []]
    annotationDict2 = annotationExtractor.getAnnotationSystem()
    error = 0
    totalSet = set()

    for entry in annotationDict1:
        if not set(
            [x for x in annotationDict2[entry]["BQB_HAS_PART"] if "uniprot" in x]
        ).issubset(
            set([x for x in annotationDict1[entry]["BQB_HAS_PART"] if "uniprot" in x])
        ) and not set(
            [x for x in annotationDict2[entry]["BQB_HAS_PART"] if "uniprot" in x]
        ).issubset(
            set(
                [x for x in annotationDict1[entry]["BQB_HAS_VERSION"] if "uniprot" in x]
            )
        ):
            print "--------------+"
            print entry
            difference = set(
                [x for x in annotationDict2[entry]["BQB_HAS_PART"] if "uniprot" in x]
            ).difference(
                set(
                    [
                        x
                        for x in annotationDict1[entry]["BQB_HAS_PART"]
                        if "uniprot" in x
                    ]
                )
            )
            print difference
            print annotationDict1[entry]
            print annotationDict2[entry]
            totalSet = totalSet.union(difference)
            # print set([x for x in annotationDict1[entry]['BQB_HAS_PART'] if 'uniprot' in x])

        if not set(
            [x for x in annotationDict2[entry]["BQB_HAS_VERSION"] if "uniprot" in x]
        ).issubset(
            set(
                [x for x in annotationDict1[entry]["BQB_HAS_VERSION"] if "uniprot" in x]
            )
        ) and not set(
            [x for x in annotationDict2[entry]["BQB_HAS_VERSION"] if "uniprot" in x]
        ).issubset(
            set([x for x in annotationDict1[entry]["BQB_HAS_PART"] if "uniprot" in x])
        ):
            print "--------------"
            print entry
            difference = set(
                [x for x in annotationDict2[entry]["BQB_HAS_VERSION"] if "uniprot" in x]
            ).difference(
                set(
                    [
                        x
                        for x in annotationDict1[entry]["BQB_HAS_VERSION"]
                        if "uniprot" in x
                    ]
                )
            )
            print difference
            totalSet = totalSet.union(difference)

            # print set([x for x in annotationDict1[entry]['BQB_IS_VERSION_OF'] if 'uniprot' in x])
    print totalSet


def batchAnnotationComparison(removedAnnotationsDir, referenceDir):
    referenceFiles = getFiles(referenceDir, "xml")
    progress = progressbar.ProgressBar()
    errorFiles = 0
    counter = 0
    errorList = []
    for fileIdx in progress(range(len(referenceFiles))):
        file = referenceFiles[fileIdx]
        result = annotationComparison(
            os.path.join(removedAnnotationsDir, file.split("/")[-1]), file, errorList
        )
        if result == 0:
            counter += 1
        elif result == -1:
            errorFiles += 1

    print 1 - counter * 1.0 / (
        len(referenceFiles) - errorFiles
    ), counter, errorFiles, len(referenceFiles)
    return errorList


if __name__ == "__main__":

    batch = False

    if batch:
        errorList = batchAnnotationComparison(
            "annotationsExpanded2", "../XMLExamples/curated"
        )

        print errorList
        print len(errorList)
    else:
        annotationFileComparison(
            "annotationsExpanded2/BIOMD0000000022.xml",
            "/home/proto/workspace/RuleWorld/atomizer/XMLExamples/curated/BIOMD0000000022.xml",
        )

    # significanceTreshold = 0.3
    # errorLogDict = {}
    # for fileName in errorList:
    #     errorLogDict[fileName] = cal.processLogFile('../curated/{0}.bngl.log'.format(fileName))
    # modelCharacterization = componentAnalysis('../curated')
    # errorResults = [(x,errorLogDict[x], modelCharacterization['curated/{0}.xml'.format(x)]) for x in errorList if 'curated/{0}.xml'.format(x)
    #        in modelCharacterization if 'forcedModification' not in errorLogDict[x] or errorLogDict[x]['forcedModification'] < significanceTreshold]
    # pprint.pprint(errorResults)
    # print len(errorResults)
    # parser = defineConsole()
    # namespace = parser.parse_args()

    # annotationFileComparison('annotationsExpanded2/BIOMD0000000048.xml', '/home/proto/workspace/RuleWorld/atomizer/XMLExamples/curated/BIOMD0000000048.xml')



from pyparsing import Word, Suppress, Optional, alphanums, Group, ZeroOrMore
from bionetgen.atomizer.utils.util import pmemoize as memoize


class CycleError(Exception):

    """Exception raised for errors in the input.

    Attributes:
        expr -- input expression in which the error occurred
        msg  -- explanation of the error
    """

    def __init__(self, memory):
        self.memory = memory


def addToDependencyGraph(dependencyGraph, label, value):
    if label not in dependencyGraph:
        dependencyGraph[label] = []
    if value not in dependencyGraph[label] and value != []:
        dependencyGraph[label].append(value)


def getURIFromSBML(moleculeName, parser, filterString=None):
    """
    filters a list of URI's so that we get only filterString  ID's
    """
    annotationList = []
    if parser:
        annotations = parser.getSpeciesAnnotation()
        if annotations[moleculeName]:
            for annotation in annotations[moleculeName]:
                annotationList.extend(getAnnotations(annotation))
    if filterString:
        annotationList = [
            x
            for x in annotationList
            if any(filterstr in x for filterstr in filterString)
        ]

    return annotationList


def addAssumptions(assumptionType, assumption, assumptions):
    assumptions[assumptionType].add(assumption)


speciesNameGrammar = (
    Word(alphanums + "_" + ":#-")
    + Suppress("()")
    + Optional(Suppress("@" + Word(alphanums + "_-")))
) + ZeroOrMore(
    Suppress("+")
    + Word(alphanums + "_" + ":#-")
    + Suppress("()")
    + Optional(Suppress("@" + Word(alphanums + "_-")))
)

nameGrammar = Word(alphanums + "_-") + ":"

rateGrammar = Word("-" + alphanums + "()")

grammar = Suppress(Optional(nameGrammar)) + (
    (Group(speciesNameGrammar) | "0")
    + Suppress(Optional("<") + "->")
    + (Group(speciesNameGrammar) | "0")
    + Suppress(rateGrammar)
) ^ (speciesNameGrammar + Suppress(Optional("<") + "->") + Suppress(rateGrammar))


@memoize
def parseReactions(reaction):
    """
    given a reaction string definition it separates the elements into reactants and products
    >>> parseReactions('A() + B() -> C() k1()')
    [['A', 'B'], ['C']]
    >>> parseReactions('A()@EC + B()@PM -> C()@PM k1()')
    [['A', 'B'], ['C']]
    >>> parseReactions('0 -> A() k1()')
    ['0', ['A']]
    """
    result = grammar.parseString(reaction).asList()
    if len(result) < 2:
        result = [result, []]
    if "<->" in reaction and len(result[0]) == 1 and len(result[1]) == 2:
        result2 = [result[1], result[0]]
        result = result2
    return result


def getAnnotations(annotation):
    """
    parses a libsbml.XMLAttributes annotation object into a list of annotations
    """
    annotationDictionary = []
    if annotation == [] or annotation is None:
        return []
    for index in range(0, annotation.getNumAttributes()):
        annotationDictionary.append(annotation.getValue(index))
    return annotationDictionary



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 14 18:17:20 2014

@author: proto
"""

import progressbar

import libsbml
from util import logMess
from sbml2bngl import SBML2BNGL as SBML2BNGL
import structures
import atomizer.moleculeCreation as mc
import os
from subprocess import call
import tempfile
import sys

# sys.path.insert(0, '../utils/')
import consoleCommands
import readBNGXML
import argparse

bioqual = [
    "BQB_IS",
    "BQB_HAS_PART",
    "BQB_IS_PART_OF",
    "BQB_IS_VERSION_OF",
    "BQB_HAS_VERSION",
    "BQB_IS_HOMOLOG_TO",
    "BQB_IS_DESCRIBED_BY",
    "BQB_IS_ENCODED_BY",
    "BQB_ENCODES",
    "BQB_OCCURS_IN",
    "BQB_HAS_PROPERTY",
    "BQB_IS_PROPERTY_OF",
    "BQB_HAS_TAXON",
    "BQB_UNKNOWN",
]

modqual = [
    "BQM_IS",
    "BQM_IS_DESCRIBED_BY",
    "BQM_IS_DERIVED_FROM",
    "BQM_IS_INSTANCE_OF",
    "BQM_HAS_INSTANCE",
    "BQM_UNKNOWN",
]

import fnmatch
import argparse


def getFiles(directory, extension):
    """
    Gets a list of <*.extension> files. include subdirectories and return the absolute
    path. also sorts by size.
    """
    matches = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, "*.{0}".format(extension)):
            matches.append(
                [
                    os.path.join(os.path.abspath(root), filename),
                    os.path.getsize(os.path.join(root, filename)),
                ]
            )

    # sort by size
    matches.sort(key=lambda filename: filename[1], reverse=False)

    matches = [x[0] for x in matches]

    return matches


from collections import defaultdict
import re


def standardizeName(name):
    """
    Remove stuff not used by bngl
    """
    name2 = name

    sbml2BnglTranslationDict = {
        "^": "",
        "'": "",
        "*": "m",
        " ": "_",
        "#": "sh",
        ":": "_",
        "α": "a",
        "β": "b",
        "γ": "g",
        " ": "",
        "+": "pl",
        "/": "_",
        ":": "_",
        "-": "_",
        ".": "_",
        "?": "unkn",
        ",": "_",
        "(": "",
        ")": "",
        "[": "",
        "]": "",
        # "(": "__",
        # ")": "__",
        # "[": "__",
        # "]": "__",
        ">": "_",
        "<": "_",
    }

    for element in sbml2BnglTranslationDict:
        name = name.replace(element, sbml2BnglTranslationDict[element])
    name = re.sub("[\W]", "", name)
    return name


def parseAnnotation(annotation):
    speciesAnnotationDict = defaultdict(list)
    lista = libsbml.CVTermList()
    libsbml.RDFAnnotationParser.parseRDFAnnotation(annotation, lista)
    for idx in range(0, lista.getSize()):
        for idx2 in range(0, lista.get(idx).getResources().getLength()):
            resource = lista.get(idx).getResources().getValue(idx2)

            qualifierType = lista.get(idx).getQualifierType()
            qualifierDescription = (
                bioqual[lista.get(idx).getBiologicalQualifierType()]
                if qualifierType
                else modqual[lista.get(idx).getModelQualifierType()]
            )
            speciesAnnotationDict[qualifierDescription].append(resource)
    return speciesAnnotationDict


def buildAnnotationDict(document):
    annotationDict = defaultdict(lambda: defaultdict(list))
    speciesNameDict = {}
    for species in document.getModel().getListOfSpecies():
        annotation = species.getAnnotation()
        transformedName = standardizeName(species.getName())
        speciesNameDict[transformedName] = species.getName()
        speciesNameDict[species.getName()] = transformedName
        if annotation:
            annotationDict[transformedName] = parseAnnotation(annotation)
    return annotationDict, speciesNameDict


def updateFromParent(child, parent, annotationDict):
    for annotationLabel in annotationDict[parent]:
        if annotationLabel in ["BQB_IS_VERSION_OF", "BQB_IS"]:
            annotationDict[child]["BQB_IS_VERSION_OF"] = annotationDict[parent][
                annotationLabel
            ]


def updateFromChild(parent, child, annotationDict):
    for annotationLabel in annotationDict[child]:
        if annotationLabel in ["BQB_IS_VERSION_OF", "BQB_IS"]:
            annotationDict[parent]["BQB_HAS_VERSION"] = annotationDict[child][
                annotationLabel
            ]


def updateFromComplex(complexMolecule, sct, annotationDict, annotationToSpeciesDict):
    localSpeciesDict = {}
    unmatchedReactants = []
    unmatchedAnnotations = []
    for constituentElement in sct[complexMolecule][0]:
        flag = False
        if len(annotationDict[constituentElement]) > 0:
            for annotation in annotationDict[constituentElement]:
                if annotation in ["BQB_IS_VERSION_OF", "BQB_IS", "BQB_HAS_VERSION"]:
                    flag = True
                    for individualAnnotation in annotationDict[constituentElement][
                        annotation
                    ]:
                        localSpeciesDict[individualAnnotation] = constituentElement
                        localSpeciesDict[constituentElement] = individualAnnotation
            if flag:
                continue

        if constituentElement in annotationToSpeciesDict:
            localSpeciesDict[constituentElement] = annotationToSpeciesDict[
                constituentElement
            ]
            localSpeciesDict[
                annotationToSpeciesDict[constituentElement]
            ] = constituentElement
        else:
            unmatchedReactants.append(constituentElement)

    for annotationType in annotationDict[complexMolecule]:
        if annotationType in ["BQB_HAS_VERSION", "BQB_HAS_PART"]:
            for constituentAnnotation in annotationDict[complexMolecule][
                annotationType
            ]:
                if constituentAnnotation not in localSpeciesDict:
                    unmatchedAnnotations.append(constituentAnnotation)
    if len(set(unmatchedReactants)) == 1 and len(set(unmatchedAnnotations)) == 1:
        localSpeciesDict[unmatchedReactants[0]] = unmatchedAnnotations[0]
        localSpeciesDict[unmatchedAnnotations[0]] = unmatchedReactants[0]
        annotationDict[unmatchedReactants[0]]["BQB_IS_VERSION_OF"] = [
            unmatchedAnnotations[0]
        ]

    elif len(unmatchedReactants) > 0 or len(unmatchedAnnotations) > 0:
        # annotate from database names
        print "**//", complexMolecule, unmatchedReactants, unmatchedAnnotations

    for element in localSpeciesDict:
        if element not in annotationToSpeciesDict:
            annotationToSpeciesDict[element] = localSpeciesDict[element]


def updateFromComponents(complexMolecule, sct, annotationDict, annotationToSpeciesDict):
    localSpeciesDict = {}
    unmatchedReactants = []
    for constituentElement in sct[complexMolecule][0]:
        flag = False
        if len(annotationDict[constituentElement]) > 0:
            for annotation in annotationDict[constituentElement]:
                if annotation in [
                    "BQB_IS_VERSION_OF",
                    "BQB_IS",
                    "BQB_HAS_VERSION",
                    "BQB_HAS_PART",
                ]:
                    for individualAnnotation in annotationDict[constituentElement][
                        annotation
                    ]:
                        # localSpeciesDict[individualAnnotation] = constituentElement
                        localSpeciesDict[constituentElement] = individualAnnotation
                        flag = True
        if not flag:
            unmatchedReactants.append(constituentElement)
    for element in localSpeciesDict:
        annotationDict[complexMolecule]["BQB_HAS_PART"].append(
            localSpeciesDict[element]
        )


def buildAnnotationTree(annotationDict, sct, database):
    """
    Deletes annotations from elements that are not molecule types
    does not remove annotations from elements that are child that did not exist in the original sbml model
    """
    for element in reversed(database.weights):
        if len(sct[element[0]]) > 0:
            if len(sct[element[0]][0]) == 1:
                if sct[element[0]][0][0] not in annotationDict:
                    continue
                if len(annotationDict[sct[element[0]][0][0]]) == 0:
                    continue
            annotationDict[element[0]] = {}


def speciesAnnotationsToSBML(sbmlDocument, annotationDict, speciesNameDict):
    """
    Receives a series of annotations associated with their associated species
    and fills in a corresponding sbmlDocument with this information
    """
    for species in sbmlDocument.getModel().getListOfSpecies():
        transformedName = speciesNameDict[species.getName()]
        if len(annotationDict[transformedName]) == 0:
            species.unsetCVTerms()
            continue
        for element in annotationDict[transformedName]:
            term = libsbml.CVTerm()
            if element.startswith("BQB"):
                term.setQualifierType(libsbml.BIOLOGICAL_QUALIFIER)
                term.setBiologicalQualifierType(bioqual.index(element))
            else:
                term.setQualifierType(libsbml.MODEL_QUALIFIER)
                term.setModelQualifierType(modqual.index(element))
            for annotation in annotationDict[transformedName][element]:
                term.addResource(annotation)
            species.addCVTerm(term)

        annotation = libsbml.RDFAnnotationParser.createAnnotation()
        cvterms = libsbml.RDFAnnotationParser.createCVTerms(species)
        rdfAnnotation = libsbml.RDFAnnotationParser.createRDFAnnotation()
        if cvterms:
            rdfAnnotation.addChild(cvterms)
        else:
            print species
        annotation.addChild(rdfAnnotation)
        species.setAnnotation(annotation)


def obtainSCT(fileName, reactionDefinitions, useID, namingConventions):
    """
    one of the library's main entry methods. Process data from a file
    to obtain the species composition table, a dictionary describing
    the chemical history of different elements in the system
    """
    logMess.log = []
    logMess.counter = -1
    reader = libsbml.SBMLReader()
    document = reader.readSBMLFromFile(fileName)

    parser = SBML2BNGL(document.getModel(), useID)
    database = structures.Databases()
    database.forceModificationFlag = True
    database = mc.createSpeciesCompositionGraph(
        parser,
        database,
        reactionDefinitions,
        namingConventions,
        speciesEquivalences=None,
        bioGridFlag=False,
    )

    return database.prunnedDependencyGraph, database, document


import tempfile


def writeSBML(document, fileName):
    writer = libsbml.SBMLWriter()
    writer.writeSBMLToFile(document, fileName)


def reduceAnnotations(fileName):

    sct, database, sbmlDocument = obtainSCT(
        fileName,
        "config/reactionDefinitions.json",
        False,
        "config/namingConventions.json",
    )
    annotationDict, speciesNameDict = buildAnnotationDict(sbmlDocument)
    buildAnnotationTree(annotationDict, sct, database)
    speciesAnnotationsToSBML(sbmlDocument, annotationDict, speciesNameDict)

    writer = libsbml.SBMLWriter()
    return writer.writeSBMLToString(sbmlDocument)


def defineConsole():
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument(
        "-i", "--input-file", type=str, help="input SBML file", required=True
    )
    parser.add_argument(
        "-o", "--output-file", type=str, help="output SBML file", required=True
    )
    return parser


def batchDeletionProcess(directory, outputDir):
    testFiles = getFiles(directory, "xml")
    progress = progressbar.ProgressBar()

    for fileIdx in progress(range(len(testFiles))):
        file = testFiles[fileIdx]
        print file
        if file in [
            "/home/proto/workspace/RuleWorld/atomizer/XMLExamples/curated/BIOMD0000000255.xml"
        ]:
            continue
        sbmlInfo = reduceAnnotations(file)
        outputFile = os.path.join(outputDir, file.split("/")[-1])

        with open(outputFile, "w") as f:
            f.write(sbmlInfo)


if __name__ == "__main__":
    # sbmlInfo = reduceAnnotations('../XMLExamples/curated/BIOMD0000000412.xml')
    # with open('annotationsRemoved/BIOMD0000000074.xml','w') as f:
    #    f.write(sbmlInfo)

    batchDeletionProcess("../XMLExamples/curated", "annotationsRemoved")
    # parser = defineConsole()
    # namespace = parser.parse_args()
    # input_file = '/home/proto/workspace/bionetgen/parsers/SBMLparser/XMLExamples/curated/BIOMD%010i.xml' % 19
    # expandedString = reduceAnnotations(namespace.input_file)
    # print 'Writing extended annotation SBML to {0}'.format(namespace.output_file)
    # with open(namespace.output_file,'w') as f:
    #    f.write(expandedString)
    # outputFileName = '.'.join(fileName.split('.')[0:-1]) + '_withAnnotations.xml'
    # writeSBML(sbmlDocument,outputFileName)



import sqlite3
import utils.annotationExtractor
import utils.annotationResolver
import argparse
import os
from copy import copy
import pprint
import fnmatch
import progressbar
import enum
import utils.readBNGXML

Query = enum.Enum("Query", "all organism species family")
organismFamilies = {
    "mammals": [
        "Homo sapiens",
        "Mammalia",
        "Mus musculus",
        "Rattus norvegicus",
        "Rattus rattus",
    ]
}


def getFiles(directory, extension):
    """
    Gets a list of <*.extension> files. include subdirectories and return the absolute
    path. also sorts by size.
    """
    matches = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, "*.{0}".format(extension)):
            matches.append(
                [
                    os.path.join(root, filename),
                    os.path.getsize(os.path.join(root, filename)),
                ]
            )

    # sort by size
    matches.sort(key=lambda filename: filename[1], reverse=False)
    matches = [x[0] for x in matches]

    return matches


class NamingDatabase:
    def __init__(self, databaseName):
        self.databaseName = databaseName

    def getAnnotationsFromSpecies(self, speciesName):
        connection = sqlite3.connect(self.databaseName)
        cursor = connection.cursor()
        queryStatement = 'SELECT annotationURI,annotationName from moleculeNames as M join identifier as I ON M.ROWID == I.speciesID join annotation as A on A.ROWID == I.annotationID and M.name == "{0}"'.format(
            speciesName
        )
        queryResult = [x[0] for x in cursor.execute(queryStatement)]
        connection.close()
        return queryResult

    def getFileNameFromSpecies(self, speciesName):
        """
        species name refers to a molecular species
        """
        connection = sqlite3.connect(self.databaseName)
        cursor = connection.cursor()
        queryStatement = 'SELECT B.file,M.name from moleculeNames as M join biomodels as B on B.ROWID == M.fileID WHERE M.name == "{0}"'.format(
            speciesName
        )
        queryResult = [x[0] for x in cursor.execute(queryStatement)]
        connection.close()
        return queryResult

    def getFileNameFromOrganism(self, organismName):
        """
        pass
        """
        connection = sqlite3.connect(self.databaseName)
        cursor = connection.cursor()
        queryStatement = 'SELECT B.file,A.annotationName from biomodels as B join annotation as A on B.organismID == A.ROWID WHERE A.annotationName == "{0}"'.format(
            organismName
        )
        queryResult = [x[0] for x in cursor.execute(queryStatement)]
        connection.close()
        return queryResult

    def getOrganismNames(self):
        connection = sqlite3.connect(self.databaseName)
        cursor = connection.cursor()
        queryStatement = "SELECT DISTINCT A.annotationName from biomodels as B join annotation as A on B.organismID == A.ROWID"
        queryResult = [x[0] for x in cursor.execute(queryStatement)]
        connection.close()
        return queryResult

    def getSpeciesFromAnnotations(self, annotation):
        connection = sqlite3.connect(self.databaseName)
        cursor = connection.cursor()
        queryStatement = 'SELECT name,A.annotationURI from moleculeNames as M join identifier as I ON M.ROWID == I.speciesID join annotation as A on A.ROWID == I.annotationID and A.annotationURI == "{0}"'.format(
            annotation
        )
        queryResult = [x[0] for x in cursor.execute(queryStatement)]
        connection.close()
        return queryResult

    def getFilesInDatabase(self):
        connection = sqlite3.connect(self.databaseName)
        cursor = connection.cursor()
        queryStatement = "SELECT file from biomodels"
        queryResult = [x[0] for x in cursor.execute(queryStatement)]
        connection.close()
        return queryResult

    def getSpeciesFromFileName(self, fileName):
        connection = sqlite3.connect(self.databaseName)
        cursor = connection.cursor()
        queryStatement = 'SELECT B.file,name,A.annotationURI,A.annotationName,qualifier from moleculeNames as M join identifier as I ON M.ROWID == I.speciesID \
                            join annotation as A on A.ROWID == I.annotationID join biomodels as B on B.ROWID == M.fileID and B.file == "{0}"'.format(
            fileName
        )

        # I.qualifier != "BQB_HAS_PART" and \
        # I.qualifier != "BQB_HAS_VERSION" AND I.qualifier != "BQB_HAS_PROPERTY"'.format(fileName)

        speciesList = [x[1:] for x in cursor.execute(queryStatement)]

        tmp = {x[0]: set([]) for x in speciesList}
        tmp2 = {x[0]: set([]) for x in speciesList}
        tmp3 = {x[0]: set([]) for x in speciesList}
        tmp4 = {x[0]: set([]) for x in speciesList}
        for x in speciesList:
            if x[3] in ["BQB_IS", "BQM_IS", "BQB_IS_VERSION_OF"]:
                tmp[x[0]].add(x[1])
                if x[2] != "":
                    tmp2[x[0]].add(x[2])
                tmp3[x[0]].add(x[3])
            else:
                tmp4[x[0]].add((x[1], x[3]))

        tmp = [
            {
                "name": set([x]),
                "annotation": set(tmp[x]),
                "annotationName": set(tmp2[x]),
                "fileName": set([fileName]),
                "qualifier": tmp3[x],
                "otherAnnotation": [tmp4[x]] if tmp4[x] else [],
            }
            for x in tmp
        ]
        return tmp

    def findOverlappingNamespace(self, fileList):
        fileSpecies = []
        if len(fileList) == 0:
            fileList = self.getFilesInDatabase()

        progress = progressbar.ProgressBar(maxval=len(fileList)).start()

        for idx in progress(range(len(fileList))):
            fileSpecies.extend(self.getSpeciesFromFileName(fileList[idx]))

        changeFlag = True
        fileSpeciesCopy = copy(fileSpecies)
        print "finished processing files, obtained {0} groups".format(len(fileSpecies))
        # progress = progressbar.ProgressBar(maxval=len((fileSpecies - 1) * (fileSpecies -1))).start()
        while changeFlag:
            try:
                changeFlag = False
                for idx in range(0, len(fileSpecies) - 1):
                    for idx2 in range(idx + 1, len(fileSpecies)):
                        # if (len(fileSpecies[idx]['name']) > 2 and fileSpecies[idx]['name'].intersection(fileSpecies[idx2]['name'])) \
                        if fileSpecies[idx]["annotation"].intersection(
                            fileSpecies[idx2]["annotation"]
                        ) or fileSpecies[idx]["annotationName"].intersection(
                            fileSpecies[idx2]["annotationName"]
                        ):

                            # print 'hello',fileSpecies[idx]['annotationName'],fileSpecies[idx2]['annotationName']
                            fileSpeciesCopy[idx]["name"] = fileSpeciesCopy[idx][
                                "name"
                            ].union(fileSpeciesCopy[idx2]["name"])
                            fileSpeciesCopy[idx]["annotation"] = fileSpeciesCopy[idx][
                                "annotation"
                            ].union(fileSpeciesCopy[idx2]["annotation"])
                            fileSpeciesCopy[idx]["annotationName"] = fileSpeciesCopy[
                                idx
                            ]["annotationName"].union(
                                fileSpeciesCopy[idx2]["annotationName"]
                            )
                            fileSpeciesCopy[idx]["fileName"] = fileSpeciesCopy[idx][
                                "fileName"
                            ].union(fileSpeciesCopy[idx2]["fileName"])
                            fileSpeciesCopy[idx]["qualifier"] = fileSpeciesCopy[idx][
                                "qualifier"
                            ].union(fileSpeciesCopy[idx2]["qualifier"])
                            fileSpeciesCopy[idx]["otherAnnotation"].extend(
                                fileSpeciesCopy[idx2]["otherAnnotation"]
                            )
                            del fileSpeciesCopy[idx2]
                            fileSpecies = fileSpeciesCopy
                            raise IOError
            except IOError:
                changeFlag = True
                continue
        # fileSpecies = [[x['name'], len(x['fileName'])] for x in fileSpecies]
        fileSpecies.sort(key=lambda x: len(x["fileName"]), reverse=True)

        # import pickle

        # with open('results.dump','wb') as f:
        #    pickle.dump(fileSpecies,f)

        return fileSpecies

    def isFileInDatabase(self, fileName):
        return isFileInDatabase(self.databaseName, fileName)


def isFileInDatabase(databaseName, fileName):
    connection = sqlite3.connect(databaseName)
    cursor = connection.cursor()
    queryStatement = 'select file from biomodels WHERE file == "{0}"'.format(fileName)
    matchingFileNames = [x[0] for x in cursor.execute(queryStatement)]
    connection.close()
    return len(matchingFileNames) > 0


def setupDatabase(databaseName):
    connection = sqlite3.connect(databaseName)
    cursor = connection.cursor()
    cursor.execute(
        """CREATE TABLE biomodels(file UNIQUE, organismID INT, FOREIGN KEY(organismID) REFERENCES annotation(ROWID))"""
    )
    cursor.execute(
        """CREATE TABLE moleculeNames(fileId INT,name, FOREIGN KEY(fileID) REFERENCES biomodels(file))"""
    )
    cursor.execute(
        """CREATE TABLE annotation(annotationURI UNIQUE ON CONFLICT IGNORE,annotationName)"""
    )
    cursor.execute(
        """CREATE TABLE identifier(annotationID INT, qualifier, speciesID INT, FOREIGN KEY(speciesID) \
                   REFERENCES moleculeName(ROWID), FOREIGN KEY(annotationID) references annotation(ROWID))"""
    )
    cursor.execute(
        """CREATE TABLE bond(fileName INT, moleculeID1 INT, moleculeID2 INT, FOREIGN KEY(fileName) REFERENCES biomodels(ROWID), \
                      FOREIGN KEY(moleculeID1) REFERENCES moleculeNames(ROWID), FOREIGN KEY(moleculeID2) REFERENCES moleculeNames(ROWID))"""
    )
    connection.commit()
    connection.close()


def extractBasicAnnotations(fileName, userDefinitions=None):
    annotations = utils.annotationExtractor.AnnotationExtractor(
        fileName, userDefinitions
    )
    elementalMolecules = [x for x in annotations.sct if annotations.sct[x] == []]
    speciesAnnotations = {
        x: annotations.getAnnotationSystem()[x] for x in elementalMolecules
    }
    modelAnnotations = annotations.getModelAnnotations()
    return speciesAnnotations, modelAnnotations


def populateDatabaseFromFile(fileName, databaseName, userDefinitions=None):
    """
    Insert annotations from file <fileName>
    """

    fileName2 = fileName.split(os.sep)[-1]
    if isFileInDatabase(databaseName, fileName2):
        print ("Database already contains annotation from file {0}".format(fileName2))
        return -1

    connection = sqlite3.connect(databaseName)
    cursor = connection.cursor()

    basicModelAnnotations, generalModelAnnotations = extractBasicAnnotations(
        fileName, userDefinitions
    )

    moleculeNames = []
    annotationNames = []
    moleculeAnnotations = []

    modelSpecies = ""
    # insert model description
    for annotation in generalModelAnnotations:
        if "taxonomy" in annotation[1]:
            modelSpecies = annotation
            break

    annotationNames.append(utils.annotationResolver.resolveAnnotation(annotation[1]))

    cursor.executemany(
        "INSERT into annotation(annotationURI,annotationName) values (?, ?)",
        annotationNames,
    )

    connection.commit()
    annotationID = [
        x
        for x in cursor.execute(
            'select ROWID from annotation WHERE annotationURI == "{0}"'.format(
                annotationNames[-1][0]
            )
        )
    ][0][0]
    annotationNames = []
    cursor.executemany(
        "INSERT into biomodels(file,organismID) values (?,?)",
        [[fileName2, annotationID]],
    )
    connection.commit()

    modelID = [
        x
        for x in cursor.execute(
            'select ROWID from biomodels WHERE file == "{0}"'.format(fileName2)
        )
    ][0][0]

    # insert moleculeNames
    for molecule in basicModelAnnotations:
        moleculeNames.append([modelID, molecule])

    annotationIDs = {
        x[1]: x[0] for x in cursor.execute("select ROWID,annotationURI from annotation")
    }

    # insert annotations
    for molecule in basicModelAnnotations:
        for annotationType in basicModelAnnotations[molecule]:
            for annotation in basicModelAnnotations[molecule][annotationType]:
                if annotation not in annotationIDs:
                    annotationName = utils.annotationResolver.resolveAnnotation(
                        annotation
                    )
                    annotationNames.append([annotation, annotationName[1]])

    cursor.executemany(
        "INSERT into annotation(annotationURI,annotationName) values (?, ?)",
        annotationNames,
    )
    connection.commit()
    cursor.executemany(
        "INSERT into moleculeNames(fileId,name) values (?, ?)", moleculeNames
    )
    connection.commit()

    moleculeIDs = {
        x[1]: x[0]
        for x in cursor.execute(
            "select ROWID,name from moleculeNames WHERE moleculeNames.fileId == '{0}'".format(
                modelID
            )
        )
    }
    annotationIDs = {
        x[1]: x[0] for x in cursor.execute("select ROWID,annotationURI from annotation")
    }

    for molecule in basicModelAnnotations:
        for annotationType in basicModelAnnotations[molecule]:
            for annotation in basicModelAnnotations[molecule][annotationType]:
                moleculeAnnotations.append(
                    [annotationIDs[annotation], moleculeIDs[molecule], annotationType]
                )
    cursor.executemany(
        "insert into identifier(annotationID,speciesID, qualifier) values (?, ?, ?)",
        moleculeAnnotations,
    )
    connection.commit()
    connection.close()


def populateBondDatabaseFromFile(fileName, databaseName):
    pass


def defineConsole():
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument(
        "-c", "--create", action="store_true", help="Create database tables"
    )
    parser.add_argument(
        "-d", "--database", type=str, help="database to modify", required=True
    )
    parser.add_argument("-i", "--input-file", type=str, help="input SBML file")
    parser.add_argument(
        "-u",
        "--user_conventions",
        type=str,
        help="Use user convention definitions for SCT calculation",
    )
    parser.add_argument(
        "-q", "--query", type=str, help="Query a database for its common namespace"
    )
    parser.add_argument(
        "-s",
        "--specific_query",
        type=str,
        help="search for models with a given molecule",
    )
    parser.add_argument(
        "-r",
        "--directory",
        type=str,
        help='Add SBML models in directory "directory" to database',
    )
    # parser.add_argument('-o','--output-file',type=str,help='output SBML file',required=True)
    return parser


def query(database, queryType, queryOptions):
    db = NamingDatabase(database)
    # db.getAnnotationsFromSpecies('EGFR')
    # db.getSpeciesFromAnnotations('http://identifiers.org/uniprot/P00533')
    # print db.getSpeciesFromFileName('BIOMD0000000048.xml')
    try:
        if Query[queryType] == Query.species:
            if queryOptions is None:
                print "Species query must indicate a species to search for using the '-s' flag"
                return
            selectedFiles = db.getFileNameFromSpecies(queryOptions)
            result = db.findOverlappingNamespace(selectedFiles)
        elif Query[queryType] == Query.organism:
            if queryOptions is not None:
                selectedFiles = db.getFileNameFromOrganism(queryOptions)
                if selectedFiles != []:
                    result = db.findOverlappingNamespace(selectedFiles)
                else:
                    result = []
            else:
                result = {}
                organismNames = db.getOrganismNames()
                for organism in organismNames:
                    selectedFiles = db.getFileNameFromOrganism(organism)
                    print organism, len(selectedFiles)
                    result[organism] = db.findOverlappingNamespace(selectedFiles)
        elif Query[queryType] == Query.family:
            result = {}
            if queryOptions is None:
                selectedFiles = []
                for family in organismFamilies:
                    selectedFiles.extend(db.getFileNameFromSpecies(family))
                    result["mammalsFamily"] = db.findOverlappingNamespace(selectedFiles)

        elif Query[queryType] == Query.all:
            result = db.findOverlappingNamespace([])
        # pprint.pprint([[x['name'], len(x['fileName'])] for x in result])
        import pickle

        with open("results2.dump", "wb") as f:
            pickle.dump(result, f)
    except KeyError:
        print "Query operation not supported"


if __name__ == "__main__":
    parser = defineConsole()
    namespace = parser.parse_args()
    if namespace.query:
        query(namespace.database, namespace.query, namespace.specific_query)
    else:
        if namespace.create:
            setupDatabase(namespace.database)

        if namespace.input_file:
            userDefinitions = namespace.user_conventions
            populateDatabaseFromFile(
                namespace.input_file, namespace.database, userDefinitions
            )
        if namespace.directory:
            sbmlFiles = getFiles(namespace.directory, "xml")
            progress = progressbar.ProgressBar(maxval=len(sbmlFiles)).start()
            for idx in progress(range(len(sbmlFiles))):
                populateDatabaseFromFile(sbmlFiles[idx], namespace.database, None)



# -*- coding: utf-8 -*-
"""
Created on Mon Sep  2 18:11:35 2013

@author: proto
"""

import platform

if platform.system() != "Windows":
    import pexpect
else:
    import winpexpect
import subprocess
import os

from os.path import expanduser, join

home = expanduser("~")

bngExecutable = join(home, "workspace", "RuleWorld", "bionetgen", "bng2", "BNG2.pl")


def setBngExecutable(executable):
    global bngExecutable
    bngExecutable = executable


def getBngExecutable():
    return bngExecutable


def bngl2xml(bnglFile, timeout=60):
    try:
        if platform.system() != "Windows":
            bngconsole = pexpect.spawn(
                "{0} --console".format(getBngExecutable()), timeout=timeout
            )
        else:
            bngconsole = winpexpect.winspawn(
                "perl {0} --console".format(getBngExecutable()), timeout=timeout
            )
        bngconsole.expect("BNG>")
        bngconsole.sendline("load {0}".format(bnglFile))
        bngconsole.expect("BNG>")
        bngconsole.sendline("action writeXML()")
        bngconsole.expect("BNG>")
        bngconsole.close()
    except pexpect.TIMEOUT:
        subprocess.call(["killall", "bngdev"])


def bngl2sbml(bnglFile, timeout=60):
    try:
        bngconsole = pexpect.spawn(
            "{0} --console".format(getBngExecutable()), timeout=timeout
        )
        bngconsole.expect("BNG>")
        bngconsole.sendline("load {0}".format(bnglFile))
        bngconsole.expect("BNG>")
        bngconsole.sendline("action generate_network()")
        bngconsole.expect("BNG>")
        bngconsole.sendline("action writeSBML()")
        bngconsole.expect("BNG>")
        bngconsole.close()
    except pexpect.TIMEOUT:
        subprocess.call(["killall", "bngdev"])


def correctness(bnglFile):
    bngconsole = pexpect.spawn("{0} --console".format(getBngExecutable()))
    bngconsole.expect("BNG>")
    bngconsole.sendline("load {0}".format(bnglFile))
    bngconsole.expect("BNG>")
    output = bngconsole.before
    bngconsole.close()
    if "ERROR" in output in output:
        return False
    return True


def writeNetwork(bnglFile):
    bngconsole = pexpect.spawn("{0} --console".format(getBngExecutable()))
    bngconsole.expect("BNG>")
    bngconsole.sendline("load {0}".format(bnglFile))
    bngconsole.expect("BNG>")
    bngconsole.sendline("action generate_network()")
    bngconsole.expect("BNG>")
    bngconsole.close()


def generateGraph(bnglFile, graphType, options=[]):
    directory = os.sep.join(bnglFile.split(os.sep)[:-1])
    bngconsole = pexpect.spawn("{0} --console".format(getBngExecutable()))
    bngconsole.expect("BNG>")
    bngconsole.sendline("load {0}".format(bnglFile))
    bngconsole.expect("BNG>")
    options = ["{0}=>1".format(x) for x in options]
    options = ", ".join(options)
    if graphType == "regulatory":
        bngconsole.sendline("action visualize({{{0}}})".format(options))
    elif graphType == "contactmap":
        bngconsole.sendline('action visualize({type=>"contactmap"})')
    else:
        return False
    bngconsole.expect("BNG>")
    bngconsole.close()
    return True


if __name__ == "__main__":
    print(bngl2xml("/tmp/tmpIhm3Ej.bngl"))
